/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2003 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *			  HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.Serizawa <serizawa@sdl.hitachi.co.jp>
 * Updated by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/lkst_private.h>
#include <linux/spinlock.h>
#include <linux/utsname.h>
#include <linux/sysrq.h>
// for registering notifier (DUMP, hotplug-cpu)
#include <linux/notifier.h>
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
#include <linux/dump.h>
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/
#include "lkst_core.h"

// TODO: check architecture independency
#include <asm/uaccess.h>
#include <asm/atomic.h>


/* trace control functions */
EXPORT_SYMBOL(lkst_trc_status);	/* get status */
EXPORT_SYMBOL(lkst_trc_start);	/* set maskset to saved one.*/
EXPORT_SYMBOL(lkst_trc_stop);	/* set maskset RNOTHING and save current.*/

/* maskset control functions */
EXPORT_SYMBOL(lkst_maskset_xchg);	/* set and return previous maskset id*/
EXPORT_SYMBOL(lkst_maskset_read);	/* read a maskset */
EXPORT_SYMBOL(lkst_maskset_write);	/* write/create a maskset */
EXPORT_SYMBOL(lkst_maskset_delete);	/* delete a maskset */
EXPORT_SYMBOL(lkst_maskset_list);	/* list all maskset *id*s */
EXPORT_SYMBOL(lkst_maskset_get_id);	/* get an id of maskset from name */

/* evhandler control functions */
EXPORT_SYMBOL(lkst_evhandler_list);	/* list all event-handler *id*s */
EXPORT_SYMBOL_GPL(lkst_evhandler_ctrl);		/* send a control message to a event-handler (GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_register);	/* register a event-hander(GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_unregister);	/* unregister a event-hander(GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_get_id);	/* get a event-hander id from name (GPL ONLY)*/

/* buffer control functions */
EXPORT_SYMBOL(lkst_buffer_create);	/* create a buffer */
EXPORT_SYMBOL(lkst_buffer_delete);	/* delete a buffer */
EXPORT_SYMBOL(lkst_buffer_read);	/* read a buffer */
EXPORT_SYMBOL(lkst_buffer_shift);	/* shift(jump to a buffer specified by
					 * next field of event buffer
					 * structure) current writing buffer */
EXPORT_SYMBOL(lkst_buffer_jump);	/* jump current writing buffer to
					 * another one */


/* evhandler primitive functions */

/* get a event buffer that has specified id. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_select_buffer_id);

/* get a event buffer that is currently writing. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_select_buffer);

/* set reading process to ready to read.(and wake up at next time scheduling)*/
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_owner_wakeup);

/* log event to current writing buffer. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_entry_log);

/* log event to static event buffer. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_entry_log_static);

/* exchange current maskset for specified maskset */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_maskset_xchg);

/* SysRq function */
EXPORT_SYMBOL(lkst_buffer_dump_current_txt);

/* internal control functions (Not use without enough understanding.)*/
/* EXPORT_SYMBOL(lkst_event_type_mnemonics); */
/* EXPORT_SYMBOL(lkst_event_type_descriptions); */
/* EXPORT_SYMBOL(lkst_hook_init_module); */
/* EXPORT_SYMBOL(lkst_hook_reset_all); */

/* default event handler control function (do nothing)*/
EXPORT_SYMBOL(lkst_evhandler_default_ctrl);

/* embeded event handlers */
/* default event handler that log event to current writing buffer. */
EXPORT_SYMBOL(lkst_evhandler_default);
/* an event handler that jump to a buffer whose id is specified next
 * field of current writing buffer. */
EXPORT_SYMBOL(lkst_evhandler_buffer_shift);
/* an event handler that log event to static event buffer.*/
EXPORT_SYMBOL(lkst_evhandler_log_static);

#if DEBUG
EXPORT_SYMBOL(lkst_evhandler_exception); /* To avoid infinit-loop*/
#endif
EXPORT_SYMBOL(lkst_entry_log);

#ifdef LKST_DEBUG_MESSAGE
#define LKST_DPRINTK(fmt, args...)	\
	printk(KERN_ERR "LKST(%s): " fmt, __FUNCTION__, ## args)

#define LKST_DPRINTK_IF(expr ,fmt, args...)	 \
	({if (expr) {LKST_DPRINTK(fmt ,## args);}})
/*for debugging*/
/* Error code for lkst_evhandlerprim_maskset_xchg_inline() */
#define LKST_MASKSET_XCHG_ERRNO_SUCCESS	(0)
#define LKST_MASKSET_XCHG_ERRNO_WCID	(-1)
#define LKST_MASKSET_XCHG_ERRNO_WCPT	(-2)
#define LKST_MASKSET_XCHG_ERRNO_ECPT	(-3)
#define LKST_MASKSET_XCHG_ERRNO_ENID	(-4)
#define LKST_MASKSET_XCHG_ERRNO_ENPT	(-5)
#define LKST_MASKSET_XCHG_ERRNO_ESTAT	(-6)
#define LKST_MASKSET_XCHG_ERRNO_ELOCK	(-7)
#else
#define LKST_DPRINTK(fmt, args...)
#define LKST_DPRINTK_IF(expr ,fmt, args...)
#define LKST_MASKSET_XCHG_ERRNO_SUCCESS	0
#define LKST_MASKSET_XCHG_ERRNO_WCID	0
#define LKST_MASKSET_XCHG_ERRNO_WCPT	0
#define LKST_MASKSET_XCHG_ERRNO_ECPT	(-EBUSY )
#define LKST_MASKSET_XCHG_ERRNO_ENID	(-EINVAL)
#define LKST_MASKSET_XCHG_ERRNO_ENPT	(-EINVAL)
#define LKST_MASKSET_XCHG_ERRNO_ESTAT	(-EBUSY )
#define LKST_MASKSET_XCHG_ERRNO_ELOCK	(-EAGAIN)
#endif


#define LKST_NOT_INIT	  0 // not initialized
#define LKST_INIT_LEVEL_0 1 // stage 0 init
#define LKST_INIT_LEVEL_1 2 // stage 1 init

static volatile int lkst_init_flag = LKST_NOT_INIT; // TODO: Don't use
static size_t lkst_init_buffer_size = LKST_BUFFER_SIZE_DEFAULT;
int lkst_init_maskset = LKST_MASKSET_ID_RDEFAULT;

spinlock_t lkst_evhandler_lock = SPIN_LOCK_UNLOCKED;
spinlock_t lkst_maskset_lock = SPIN_LOCK_UNLOCKED;

/* definitions of special masksets. */
/*
 * maskset_rnothing (id: 0(LKST_MASKSET_ID_RNOTHING))
 * a maskset to record no events.
 */
static const lkst_evhandler_id maskset_rnothing[] = {
	[0 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_VOID
		/* caution: LKST_ETYPE_MAX is significant
		 *	 to decide the size of this maskset.
		 */
};

/*
 * maskset_rall (id: 1(LKST_MASKSET_ID_RALL))
 * a maskset to record all events.
 */
static lkst_evhandler_id maskset_rall[] = {
	[0 ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_BUFFER_SHIFT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
		/* caution: LKST_ETYPE_MAX is significant, ditto. */
};

/*
 * maskset_rdefault (id: 2(LKST_MASKSET_ID_RDEFAULT))
 * a maskset selected initially.
 */
	/* Default Maskset omits lock operation events */
static const lkst_evhandler_id maskset_rdefault[] = {
#ifndef CONFIG_X86_4G
	[0 ... LKST_ETYPE_MEM_DO_NOPAGE - 1 ] = LKST_EVHANDLER_ID_DEFAULT,
#else
	[0 ... LKST_ETYPE_PROCESS_CONTEXTSWITCH - 1 ] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_PROCESS_CONTEXTSWITCH ... LKST_ETYPE_INT_BH_ENTRY] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_INT_BH_ENTRY + 1 ... LKST_ETYPE_SYSCALL_ENTRY - 1] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_SYSCALL_ENTRY ... LKST_ETYPE_MEM_DO_NOPAGE-1] = LKST_EVHANDLER_ID_DEFAULT,
#endif
	[LKST_ETYPE_MEM_DO_NOPAGE ... LKST_ETYPE_NET_PKTSENDI-1 ] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_NET_PKTSENDI] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_NET_PKTRECV]  = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_NET_PKTRECV+1 ... LKST_ETYPE_LK_SPINLOCK -1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LK_SPINLOCK ... LKST_ETYPE_O_PORTOUT] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_O_PORTOUT + 1 ... LKST_ETYPE_TIMER_ADD - 1 ] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_TIMER_ADD ... LKST_ETYPE_LKST_INIT - 1 ] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_LKST_INIT ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_BUFFER_SHIFT,
#ifndef CONFIG_X86_4G
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
#else
	[LKST_ETYPE_LKST_BUFF_HIGH] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_HIGH + 1 ... LKST_ETYPE_LKST_EXTEND - 1] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_LKST_EXTEND ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
#endif
		/* caution: LKST_ETYPE_MAX is significant, ditto. */
};

/* file operators */
static int lkst_open(struct inode *inode, struct file *filp);
static int lkst_close(struct inode *inode, struct file *filp);
static ssize_t lkst_read(struct file *filp, char *buf,
			 size_t count, loff_t *ppos);
static int lkst_ioctl(struct inode *inode, struct file *filp,
		      unsigned int request, unsigned long arg);

static struct file_operations lkst_cdev_fops = {
	.owner=		THIS_MODULE,
	.ioctl=		lkst_ioctl,
	.open=		lkst_open,
	.read=		lkst_read,
	.release=	lkst_close,
};

#ifdef CONFIG_MAGIC_SYSRQ
static void sysrq_handle_lkst_dump_buffer(int key, struct pt_regs *pt_regs,
					  struct kbd_struct *kbd, 
					  struct tty_struct *tty)
{
	lkst_buffer_dump_current_txt();
}

static struct sysrq_key_op sysrq_lkst_dump_buffer_op = {
	.handler=		sysrq_handle_lkst_dump_buffer,
	.help_msg=		"lkstView",
	.action_msg=		"Dump Buffer\n",
};
#endif /* CONFIG_MAGIC_SYSRQ */

/* for devfs support */
#ifdef CONFIG_DEVFS_FS
static devfs_handle_t lkst_devfs_handle = NULL;
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
static devfs_handle_t lkst_devfs_handle_cpu[LKST_CPU_MAX] = NULL;
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
#endif

/*kernel dump notify handler */
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
static int lkst_dump_notify_handler(struct notifier_block *self,
					unsigned long stat,
					void *dev);

static struct notifier_block dump_nb = {
	.notifier_call = lkst_dump_notify_handler,
	.next = NULL,
	.priority = 0
};
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

#if DEBUG
/* enable messages for debug*/
#define LKST_DEBUG_MESSAGE
#define LKST_DEBUG_LOG(ty, a1, a2, a3, a4) \
	({lkst_evhandlerprim_entry_log_static(ty,a1,a2,a3,a4);})
#else /* DEBUG */
#define LKST_DEBUG_LOG(ty, a1, a2, a3, a4)
#endif /* DEBUG */

void lkst_entry_log(void *phookrec, int event_type,
			lkst_arg_t arg1, lkst_arg_t arg2,
			lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandler_id evhandler_id;
	volatile lkst_maskset *maskset;
	lkst_evhandler_func evhandler;

	if (!LKST_ETYPE_VERIFY(event_type))
		return;

	/*get the maskset lock*/
	if (!spin_trylock(&lkst_maskset_lock)) return;

	if (!LKST_POINTER_LOAD(lkst.current_masksetp, maskset)) {
		LKST_DEBUG_LOG(event_type + 0x200, arg1,
				   (unsigned int) (lkst.current_masksetp),
				   (unsigned int) (lkst.current_masksetp_check),
				   (unsigned int) 0);
		spin_unlock(&lkst_maskset_lock);
		return ;
	}

	evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, event_type);

	/*release the maskset lock*/
	spin_unlock(&lkst_maskset_lock);

	if (evhandler_id > LKST_EVHANDLER_ID_MAX)
		return ;

	/*get the evhandler lock*/
	if (!spin_trylock(&lkst_evhandler_lock)) return;

	if (!LKST_POINTER_LOAD
		(lkst.evhandler_table[evhandler_id].evhandler, evhandler)) {
		LKST_DEBUG_LOG(event_type + 0x400, evhandler_id,
				   (unsigned int) (evhandler),
				   (unsigned int) (lkst.evhandler_table[evhandler_id].evhandler_check),
				   (unsigned int) 0);
	} else {
		/* Avoid infinate loop. */
		if (evhandler != lkst_entry_log)
			(*evhandler) (NULL, event_type, arg1, arg2, arg3, arg4);
	}
	/*release the evhandler lock*/
	spin_unlock(&lkst_evhandler_lock);

	return;
}
/*********************************************************************
 * evhandler primitive functions
 *********************************************************************/

/*
 *	The buffer read queue mechanism :
 * A buffer writing on is in a queue tail except for a spacial case that the
 * buffer is read and not shifted-in, while reading it. When the reading
 * process begin to read a buffer, the buffer should be dequeued. And
 * read_buf field of lkst_current.owner point it. When the writing process
 * begin to write (shift-in) a buffer, the buffer should be enqueued.
 * This mechanism is adopted instead of the old wcnt/rcnt mechanism.
 * IOW, this is just a kind of LRU.(simple implementation:-) )
 */

/*
 * lkst_buffer_jump_core:
 * Jump to specified buffer from current buffer.
 */
static inline int lkst_buffer_jump_core(lkst_buffer_id next) {
	struct lkst_event_buffer *nxtp, *curp;
	unsigned long flags;
	register int ret;

	local_irq_save(flags);

	if (next == lkst_current.write_buf ||
		next == LKST_BUFFER_ID_VOID ||
		!LKST_BUFFER_LOAD(next, nxtp)) {
		/* Either next buffer does not exist, or same as current */
		ret = -EINVAL;
		goto endof_irq_saved_region;
	}
	if (!LKST_CURRENT_BUFFER_LOAD(curp)){
		ret = -EBUSY;
		goto endof_irq_saved_region;
	}

	/* LKST internal event */
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_BUFF_SHIFT,
			 LKST_ARG(lkst_current.write_buf),
			 LKST_ARG(next),
			 LKST_ARGP(curp),
			 LKST_ARGP(nxtp));

	if ( (ret = lkst_buffer_enqueue_inline(nxtp) ) == 0) {
		/* success to enqueue, then initialize */
		nxtp->write_offset = nxtp->baseid - LKST_BUFFER_INIT_POS;
		lkst_current.write_buf = next;
		lkst_current.owner.ready = 1; /* ready to read */
	}
 endof_irq_saved_region:
	local_irq_restore(flags);
	return ret;
}

/*
 * lkst_buffer_shift_core:
 * Jump to next buffer from current buffer.
 */
static inline int lkst_buffer_shift_core(void) {
	struct lkst_event_buffer * tmp;
	if ( ! LKST_CURRENT_BUFFER_LOAD(tmp) ) return -EBUSY;
	return lkst_buffer_jump_core(tmp->next);
}

/*
 * subfunctions for reading buffers
 */
/*
 * fill header of the log buffer.
 * that read from global header.
 */
static __inline__ void \
lkst_log_header_read_inline(struct lkst_log_buffer * lbuffer) {
	/* get buffer time stamp */
	lbuffer->tsc = lkst.tsc;
	lbuffer->xtime = lkst.xtime;
	lbuffer->cpu_freq = lkst.cpu_freq;
	lbuffer->buf_ver = lkst.buf_ver;
	strncpy(lbuffer->arch, lkst.arch, LKST_ARCH_NAME_LEN);
	lbuffer->endian_big = lkst.endian_big;
}


/*
 * get the next buffer from read queue and dequeue current buffer(if the
 * next buffer differ from current)
 */

static __inline__ struct lkst_event_buffer * \
lkst_buffer_get_next_readable(int overwritten ) {
	struct lkst_event_buffer *buffer;
	unsigned long flags;
	local_irq_save(flags);

	if ( lkst_current.rq_head != LKST_BUFFER_ID_VOID ) {
		/* then dequeue new buffer */
		if ( LKST_BUFFER_LOAD(lkst_current.rq_head, buffer) ) {
			if ( lkst_buffer_dequeue_inline(buffer) == 0 ) {
				goto  success;
			}
		}
	} else if ( overwritten ) {
		/* then load same buffer */
		if ( lkst_current.owner.read_buf != LKST_BUFFER_ID_VOID &&
		     LKST_BUFFER_LOAD(lkst_current.owner.read_buf, buffer) ) {
			goto  success;
		}
	}
	local_irq_restore(flags);
	return NULL;
 success:
	lkst_current.owner.read_pos = LKST_BUFFER_INIT_POS - 1;
	lkst_current.owner.read_baseid =
		LKST_BUFFER_BODY(buffer)[LKST_BUFFER_INIT_POS].log_recid - 1;
	lkst_current.owner.read_buf = buffer->id;
	local_irq_restore(flags);
	return buffer;
}

/*
 * read entrys from a buffer less than log_num, until breaking
 * baseid incrementation
 */
static int __inline__ \
lkst_buffer_entry_read_core(struct lkst_log_record * log, int log_num,
			    struct lkst_event_record * body, int size,
			    struct lkst_buffer_owner * owner,
			    int * overwritten) {
	int read_pos;
	int count, baseid, bufferid;
	struct task_struct *task = NULL;
	struct lkst_event_record *entry;
	uid_t uid = 0;
	gid_t gid = 0;
	pid_t pid = LKST_OWNER_PID_VOID;
	pid_t pgrp = 0;

	*overwritten = 0;
	read_pos = owner->read_pos;	/*CARE: read_pos might indicate negative value.*/
	baseid = owner->read_baseid;
	bufferid = owner->read_buf;
	for ( count = 0; count < log_num; count++, log++) {
		read_pos++; baseid++;
		entry = &body[(unsigned)read_pos%size]; //TODO?
		log->posix.log_pid = entry->log_pid;
		log->posix.log_time.tv_sec = (time_t) ((entry->log_time) >> 32);
		log->posix.log_time.tv_nsec = (unsigned long) entry->log_time;
		log->posix.log_thread = -1;		// TODO: write after
		log->posix.log_event_type = entry->log_event_type;
		log->log_arg1 = entry->log_arg1;
		log->log_arg2 = entry->log_arg2;
		log->log_arg3 = entry->log_arg3;
		log->log_arg4 = entry->log_arg4;
		log->posix.log_processor = (posix_log_procid_t)(entry->processor);
		if ( entry->log_recid != baseid ) {
			baseid--;
			read_pos--;
			if (body[(unsigned)read_pos%size].log_recid != baseid) {
				*overwritten = 1;
				log->posix.log_recid = LKST_BUFFER_RECID(bufferid, baseid+1);
				log->posix.log_event_type = LKST_ETYPE_LKST_OVWRTN_REC;
				log->posix.log_magic = LKST_LOG_MAGIC_NUM;
				log->posix.log_size = sizeof(lkst_arg_t) * 4;
				log->posix.log_format = PXLOG_BINARY;
				log->posix.log_facility = LOG_KERN;
				log->posix.log_severity = LOG_DEBUG;
				log->posix.log_uid = uid;
				log->posix.log_gid = gid;
				log->posix.log_pgrp = pgrp;
				log->posix.log_flags = 0;
				log++; count++;
			}
			break;
		}
		log->posix.log_recid = LKST_BUFFER_RECID(bufferid, baseid);
		read_pos = (unsigned)read_pos%size; //update read position

		/* Is the current (logged) pid same as the last?*/
		if (pid != log->posix.log_pid) {
			/*update uid, gid and pgrp.*/
			pid = log->posix.log_pid;
			read_lock(&tasklist_lock);
			task = find_task_by_pid(pid);
			if (task == NULL) {
				/* write dummy value if task_struct was already deleted */
				uid = -1;	// TODO: write after
				gid = -1;	// TODO: write after
				pgrp = -1;	// TODO: write after
			} else {
				uid = task->uid;
				gid = task->gid;
				pgrp = task->pgrp;
			}
			read_unlock(&tasklist_lock);
		}
		log->posix.log_magic = LKST_LOG_MAGIC_NUM;
		log->posix.log_size = sizeof(lkst_arg_t) * 4;
		log->posix.log_format = PXLOG_BINARY;
		log->posix.log_facility = LOG_KERN;
		log->posix.log_severity = LOG_DEBUG;
		log->posix.log_uid = uid;
		log->posix.log_gid = gid;
		log->posix.log_pgrp = pgrp;
		log->posix.log_flags = 0;
	}
	owner->read_pos = read_pos;
	owner->read_baseid = baseid;
	return count;
}

/*
 * function lkst_buffer_read_all_to_user_core
 * Read from the buffers.
 * This must be called from user space.
 *
 *  Arguments
 *    logbuffer:	specify user-space buffer.
 *    count:		number of read entries, described as entry counts. not bytes.
 *
 *  returns
 *    result read size:	the size of entries that was read.(counts)
 *    -EAGAIN:		read queue is empty. please read again after a minuts.
 *    -ENOMEM:		failed to allocate temporary buffer.
 *    -EBUSY:		buffer is currently busy.(may be or there is a bug)
 *
 */
static __inline__ \
int lkst_buffer_read_all_to_user_core(struct lkst_log_record * logbuffer,
				      int count)
{
	struct lkst_event_buffer * buffer;
	struct lkst_log_record * tmp_buf; //TODO: not vmalloc
	int log_pos, read_cnt, log_num, overwritten = 0;

	/* OK, this is *PRIVATE* function, save pointless argument check. */

	/* allocate temprary buffer */
	tmp_buf = (struct lkst_log_record *) vmalloc(LKST_TEMP_BUFFER_BYTE);
	if ( tmp_buf == NULL ) return -ENOMEM;
	log_num = LKST_TEMP_BUFFER_COUNT;
	log_pos = 0;

	/* get the oldest buffer*/
	if ( lkst_current.owner.read_buf != LKST_BUFFER_ID_VOID ) {
		if ( !LKST_BUFFER_LOAD(lkst_current.owner.read_buf, buffer)) {
			log_pos = -EBUSY;
			goto endof_temp_alloc_region;
		}
	} else {
		/* first time, or the buffer before read is deleted */
		if (! (buffer = lkst_buffer_get_next_readable(0))) {
			log_pos = -EAGAIN;
			goto endof_temp_alloc_region;
		}
	}
	/* header is needless to read */
	/* reading loop */
	do {
		if ( (count - log_pos) < log_num )
			log_num = (count - log_pos);

		read_cnt = lkst_buffer_entry_read_core(tmp_buf, log_num,
						       LKST_BUFFER_BODY(buffer),
						       buffer->size,
						       &lkst_current.owner,
						       &overwritten);
		if ( read_cnt < log_num ) {	/* end of this buffer */
			buffer = lkst_buffer_get_next_readable(overwritten);
		}
		if ( read_cnt > 0 ) {
			/* temp buffer copy to user space */
			if ( copy_to_user(
				     &logbuffer[log_pos], tmp_buf,
				     read_cnt*sizeof(struct lkst_log_record))){
				printk(KERN_ERR "Failed to copy to user space:"
				       "address %p, size " LKST_FMT_SZT "\n",
				       &logbuffer[log_pos],
				       (size_t)read_cnt*sizeof(
					       struct lkst_log_record));
				break;//TODO : return correct error code.
			}
			log_pos += read_cnt;
		}
	} while ( buffer != NULL && log_num > 0 );

 endof_temp_alloc_region:
	vfree(tmp_buf);
	return log_pos;
}

/*
 * function lkst_buffer_read_one_to_user_core
 * Read from specified buffer.
 * This must be called from user space.
 *
 *  Arguments
 *    id:		If specify LKST_BUFFER_ID_VOID, then read static-shared
 *			buffer, otherwise read specified buffer.
 *    logbuffer:	specify user-space buffer.
 *    count:		number of read *entries*, described as entry counts.
 *			Not bytes.
 *
 *  returns
 *    result read size:	the size of entries that was read.(counts)
 *    -EINVAL:		specified buffer can not use.
 *    -ENOMEM:		failed to allocate temporary buffer.
 *
 */
static __inline__ \
int lkst_buffer_read_one_to_user_core(lkst_buffer_id id,
				      struct lkst_log_record * logbuffer,
				      int count)
{
	struct lkst_buffer_owner owner;
	struct lkst_log_record * tmp_buf;
	int log_pos, read_cnt, log_num, overwritten=0;
	struct lkst_event_record *body;
	int size;

	/* allocate temprary buffer */
	tmp_buf = (struct lkst_log_record *) vmalloc(LKST_TEMP_BUFFER_BYTE);
	if ( tmp_buf == NULL ) return -ENOMEM;
	log_num = LKST_TEMP_BUFFER_COUNT;
	log_pos = 0;

	/* get buffer body and size*/
	if ( id == LKST_BUFFER_ID_VOID ) { // static shared buffer
		unsigned int start_recid;
		start_recid = (unsigned)(atomic_read(&lkst.static_buffer.recid) + 2);
		if ( start_recid < LKST_STATIC_BUFFER_COUNT )
			start_recid = 0;
		else
			start_recid %= LKST_STATIC_BUFFER_COUNT;
		body = LKST_STATIC_BUFFER_BODY();
		size = LKST_STATIC_BUFFER_COUNT;
		owner.read_pos = start_recid - 1;
		owner.read_buf = 0;
		owner.read_baseid = body[start_recid].log_recid - 1;
	} else { /* get specified buffer */  // percpu buffer
		struct lkst_event_buffer * buffer;
		if (! LKST_BUFFER_LOAD(id, buffer) ) {
			vfree(tmp_buf);
			return -EINVAL;
		}
		body = LKST_BUFFER_BODY(buffer);
		size = buffer->size;
		owner.read_pos = LKST_BUFFER_INIT_POS - 1;
		owner.read_buf = id;
		owner.read_baseid = body[LKST_BUFFER_INIT_POS].log_recid - 1;
	}

	/* reading loop */
	do {
		if ( (count - log_pos) < log_num )
			log_num = (count - log_pos);

		read_cnt = lkst_buffer_entry_read_core(tmp_buf, log_num,
						       body, size,
						       &owner,
						       &overwritten);
		if ( read_cnt > 0 ) {
			/* temp buffer copy to user space */
			if ( copy_to_user(
				     &logbuffer[log_pos], tmp_buf,
				     read_cnt*sizeof(struct lkst_log_record))){
				printk(KERN_ERR "Failed to copy to user space:"
				       "address %p, size " LKST_FMT_SZT "\n",
				       &logbuffer[log_pos],
				       (size_t)read_cnt*sizeof(
					       struct lkst_log_record));
				break;//TODO : return correct error code.
			}
			log_pos += read_cnt;
		}
	} while ( log_num > 0 && read_cnt == log_num);

	vfree(tmp_buf);
	return log_pos;
}

/*
 * function lkst_buffer_read_core
 * Read from the specified buffer.
 * This is called from lkst_buffer_read
 *
 *  Arguments
 *    lbuffer:		the logging buffer.
 *
 *  returns
 *    result read size:	the size of entries.(counts)
 *    -EINVAL:		spacified buffer is not exist.
 *
 *  !NOTE!: This must be called after CPU migration.
 *
 */
static __inline__ \
int lkst_buffer_read_core(struct lkst_log_buffer * lbuffer)
{
	struct lkst_event_buffer * buffer;
	struct lkst_buffer_owner owner;
	int ow;

	if( lbuffer->id == LKST_BUFFER_ID_VOID ||
	    ! LKST_BUFFER_LOAD(lbuffer->id, buffer) ) {
		return -EINVAL;
	}
	lbuffer->cpu = smp_processor_id();
	owner.read_pos = LKST_BUFFER_INIT_POS;
	owner.read_buf = lbuffer->id;
	owner.read_baseid =
		LKST_BUFFER_BODY(buffer)[LKST_BUFFER_INIT_POS].log_recid;
	lkst_log_header_read_inline(lbuffer);
	return lbuffer->result_read_size =
		lkst_buffer_entry_read_core(lbuffer->buffer,
					    lbuffer->read_size,
					    LKST_BUFFER_BODY(buffer),
					    buffer->size,
					    &owner, &ow);
}


/*
 * Event buffer selecting functions
 */
struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer_id(lkst_buffer_id id)
{
	struct lkst_event_buffer * buffer;
	if ( id != LKST_BUFFER_ID_VOID &&
		 LKST_BUFFER_LOAD(id, buffer) )
		return buffer;
	else
		return NULL;
}

struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer(void)
{
	struct lkst_event_buffer * buffer;
	if ( LKST_CURRENT_BUFFER_LOAD(buffer) )
		return buffer;
	else
		return NULL;
}

/*
 * Logging subfunctions are followings.
 */

/*
 * function lkst_evhandlerprim_owner_wakeup_inline
 * A primitive part of an Event-handler. This awake a owner process which
 * is waiting to read a specified buffer list.
 *
 *	Arguments
 *	  NONE
 *
 *	returns
 *	  NONE
 *
 */
static inline void lkst_evhandlerprim_owner_wakeup_inline(void)
{
	lkst_current.owner.ready = 1; /* ready to read */
}

void lkst_evhandlerprim_owner_wakeup(void)
{
	lkst_evhandlerprim_owner_wakeup_inline();
}

/*
 * function lkst_evhandlerprim_entry_log_inline
 * A primitive part of an Event-handler. This function write specified record
 * to specified entry.
 *
 *	Caution: This function does not verify value of event_entry_p.
 * TODO: remove cli ...
 * Finally, I had chosen safety rather then speed.
 * But, I would provide no-cli handler by addon-module.
 *
 *	Arguments
 *	  event_type {IN}:	corresponding event type.
 *	  arg1-4 {IN}:	arguments acquired at a trace point
 *
 *	returns
 *	  NONE
 *
 */
static inline void \
lkst_evhandlerprim_entry_log_inline(int event_type,
					lkst_arg_t arg1, lkst_arg_t arg2,
					lkst_arg_t arg3, lkst_arg_t arg4)
{
	struct lkst_event_buffer *buffer;
	int pos;
	unsigned long flags;
	int baseid;
	if (! LKST_CURRENT_BUFFER_LOAD(buffer)) return ;

	local_irq_save(flags);/* local irq disabling for atomicity */

	/* get a entry position */ {
		baseid = buffer->baseid++;
	}
	/* adjust the entry position */
	if ( (pos = baseid - buffer->write_offset) >= buffer->size ) {
		pos %= buffer->size;
	}
	/* logging */ {
		register struct lkst_event_record * event_entry_p =
			&(LKST_BUFFER_BODY(buffer)[pos]);
		event_entry_p->log_recid = baseid;
		event_entry_p->log_time = lkst_evhandlerprim_mc();
		event_entry_p->log_event_type = event_type;
		event_entry_p->log_pid = current->pid; /* current pid */ ;
		event_entry_p->log_arg1 = arg1;
		event_entry_p->log_arg2 = arg2;
		event_entry_p->log_arg3 = arg3;
		event_entry_p->log_arg4 = arg4;
	}
	/* Check buffer wrapping. */
	if ( pos == 0 ) {
		/* notify current writing buffer overflowed,
		 * occurs LKST_ETYPE_LKST_BUFF_OVFLOW event, and
		 * set reading process to ready to read. */
		buffer->write_offset += buffer->size;
		/* Wrapped. */
		/* Process buffer-over-flow event. */
		LKST_HOOK_INLINE(LKST_ETYPE_LKST_BUFF_OVFLOW,
				 LKST_ARG(smp_processor_id()),
				 LKST_ARG(buffer->next),
				 LKST_ARG(event_type),
				 LKST_ARGP(buffer));
		/* owner wakeup */
		lkst_evhandlerprim_owner_wakeup_inline();
	}
	local_irq_restore(flags);
}


/*
 * function lkst_evhandlerprim_entry_log_inline
 * A primitive part of an Event-handler. This write an event log to
 * a currently selected buffer correspond to a cpu on which the event
 * has occurred.
 *
 *	Arguments
 *	  event_type {IN}:	corresponding event type.
 *	  arg1-4 {IN}:	arguments acquired at a trace point
 *
 *	returns
 *	  NONE
 */
void lkst_evhandlerprim_entry_log(int event_type,
				  lkst_arg_t arg1, lkst_arg_t arg2,
				  lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_inline(event_type,
						arg1, arg2, arg3, arg4);
}


/*
 * function lkst_evhandlerprim_entry_log_sbuf_inline
 * A primitive part of an Event-handler. This write an event log to
 * a shared static buffer on which the event has occurred.
 *
 *	Arguments
 *	  event_type {IN}:	corresponding event type.
 *	  arg1-4 {IN}:	arguments acquired at a trace point
 *
 *	returns
 *	  NONE
 */
static inline void \
lkst_evhandlerprim_entry_log_static_inline(int event_type,
					   lkst_arg_t arg1, lkst_arg_t arg2,
					   lkst_arg_t arg3, lkst_arg_t arg4)
{
	struct lkst_event_record *event_entry_p;
	__volatile__ int recid;
	recid = atomic_read_and_add(1, &lkst.static_buffer.recid);
	event_entry_p = &LKST_STATIC_BUFFER_BODY()[
		((unsigned)recid & LKST_STATIC_BUFFER_MASK)];

	/* logging */
	event_entry_p->log_recid = recid;
	event_entry_p->processor = smp_processor_id(); /*current processor*/
	event_entry_p->log_time = lkst_evhandlerprim_mc(); /* current time */
	event_entry_p->log_event_type = event_type;
	event_entry_p->log_pid = current->pid; /* current pid */ ;
	event_entry_p->log_arg1 = arg1;
	event_entry_p->log_arg2 = arg2;
	event_entry_p->log_arg3 = arg3;
	event_entry_p->log_arg4 = arg4;
}


void lkst_evhandlerprim_entry_log_static(int event_type,
					 lkst_arg_t arg1, lkst_arg_t arg2,
					 lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_static_inline(event_type,
						   arg1, arg2, arg3, arg4);
}


/*
 * function lkst_evhandlerprim_maskset_xchg_inline
 * A primitive part of an Event-handler. This primitive
 * change current selected maskset to a specified id.
 * This primitive is for LKST internal use.
 *
 *	Arguments
 *	  id {IN}:	An id of a maskset to be selected next.
 *
 *	returns
 *	  ret:	error code (private).
 */

static inline int
lkst_evhandlerprim_maskset_xchg_inline(lkst_maskset_id * id)
{
	int	ret = LKST_MASKSET_XCHG_ERRNO_SUCCESS;
	lkst_maskset_id old_id = lkst.current_maskset_id, new_id = *id;
	volatile lkst_maskset *old_maskset = NULL, *new_maskset = NULL;

	if (new_id > LKST_MASKSET_ID_MAX) {
		/* Verify new maskset id. */
		return LKST_MASKSET_XCHG_ERRNO_ENID;
	}
	/* if we can't get maskset lock, return an error. */
	if (!spin_trylock(&lkst_maskset_lock) ) {
		return LKST_MASKSET_XCHG_ERRNO_ELOCK;
	}

 /*start of locked region */
	/* if new maskset has not been initialized, return as error. */
	if (!LKST_POINTER_LOAD(lkst.maskset_table[new_id].masksetp,
				   new_maskset)) {
		ret = LKST_MASKSET_XCHG_ERRNO_ENPT;
		goto  endof_atomic_region;
	}

	/* Load current maskset pointer. */
	if ( LKST_POINTER_LOAD(lkst.current_masksetp, old_maskset)) {
		/* Verify old maskset id. */
		if (old_maskset != lkst.maskset_table[old_id].masksetp) {
			ret = LKST_MASKSET_XCHG_ERRNO_WCPT;
			old_id = LKST_MASKSET_ID_RDEFAULT;
		} else if (old_id > LKST_MASKSET_ID_MAX) {
			ret = LKST_MASKSET_XCHG_ERRNO_WCID;
			old_id = LKST_MASKSET_ID_RDEFAULT;
		}
	} else {
		/*if current maskset == NULL*/
		ret = LKST_MASKSET_XCHG_ERRNO_WCPT;
		old_id = LKST_MASKSET_ID_RNOTHING;
	}

	/* if old maskset does not exist, return as error. */
	if ( !LKST_POINTER_LOAD(lkst.maskset_table[old_id].masksetp,
				old_maskset) ) {
		ret = LKST_MASKSET_XCHG_ERRNO_ECPT;
		goto  endof_atomic_region;
	}

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_MSET_XCHG,
			   LKST_ARG(old_id), LKST_ARG(new_id),
			   LKST_ARGP(old_maskset), LKST_ARGP(new_maskset));

	LKST_MASKSET_SET_CURRENT(new_id);

	*id = old_id;

	lkst_hook_maskset_xchg(new_id, old_id);

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_MSET_XCHG,
			   LKST_ARG(old_id), LKST_ARG(new_id),
			   LKST_ARGP(old_maskset), LKST_ARGP(new_maskset));

 endof_atomic_region:
	spin_unlock(&lkst_maskset_lock);
	return ret;
}


int lkst_evhandlerprim_maskset_xchg(lkst_maskset_id * id)
{
#ifdef LKST_DEBUG_MESSAGE
	switch(lkst_evhandlerprim_maskset_xchg_inline(id)) {
	case LKST_MASKSET_XCHG_ERRNO_SUCCESS:
	case LKST_MASKSET_XCHG_ERRNO_WCPT:
	case LKST_MASKSET_XCHG_ERRNO_WCID:
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_ECPT:
	case LKST_MASKSET_XCHG_ERRNO_ESTAT:
		return -EBUSY;
	case LKST_MASKSET_XCHG_ERRNO_ELOCK:
		return -EAGAIN;
	case LKST_MASKSET_XCHG_ERRNO_ENID:
	case LKST_MASKSET_XCHG_ERRNO_ENPT:
	default:
		return -EINVAL;
	}
#else /*LKST_DEBUG_MESSAGE*/
	return lkst_evhandlerprim_maskset_xchg_inline(id);
#endif
}

/*********************************************************************
 * Event handle functions (embeded)
 *********************************************************************/

/* default event handler function */
void lkst_evhandler_default(void *phookrec, int event_type,
				lkst_arg_t arg1,
				lkst_arg_t arg2,
				lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_inline(event_type,
						arg1, arg2, arg3, arg4);
}

/* default event handler control function */
int lkst_evhandler_default_ctrl(void *buf, size_t bufsize)
{
	return 0;
}

/* default event handler function */
void lkst_evhandler_log_static(void *phookrec, int event_type,
				   lkst_arg_t arg1,
				   lkst_arg_t arg2,
				   lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_static_inline(event_type,
						 arg1, arg2, arg3, arg4);
}

/* Event handler for buffer shifting */
void lkst_evhandler_buffer_shift(void *phookrec, int event_type,
				 lkst_arg_t arg1,
				 lkst_arg_t arg2,
				 lkst_arg_t arg3,
				 lkst_arg_t arg4)
{
	lkst_buffer_shift_core();
}

#if DEBUG
/*
 * event-handler-function lkst_evhandler_exception
 *
 * This event handler is used for exception event.
 * It can record multiple exception events, the first event is
 * written to currently selected buffer, and after second events are
 * written to static buffer.
 *
 * This event handler is used as not to crush linux kernel because of
 * occurring multiple exception event.
 *
 * "exhandler_addr" is an exception handler address.
 *
 */
void lkst_evhandler_exception(void *phookrec,
				  int event_type,
				  lkst_arg_t arg1,
				  lkst_arg_t arg2,
				  lkst_arg_t arg3,
				  lkst_arg_t arg4)
{
	static lkst_arg_t exhandler_addr = 0;
	unsigned long flags;

	local_irq_save(flags);

	/* exception handler address check */
	if (exhandler_addr != arg1) {
		exhandler_addr = arg1;
		lkst_evhandlerprim_entry_log_inline(event_type,
							arg1, arg2,
							arg3, arg4);
	} else {
		/* write to static buffer*/
		lkst_evhandlerprim_entry_log_static_inline(event_type,
							 arg1, arg2,
							 arg3, arg4);
		exhandler_addr = 0;
	}


	local_irq_restore(flags);

	return;
}
#endif /* DEBUG */

/* while dumping, stop logging.*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
static int lkst_dump_notify_handler(struct notifier_block *self,
					unsigned long stat,
					void *dev)
{
	static lkst_maskset_id maskset_id_saved = 0;
	if ( stat == DUMP_BEGIN ) {
		maskset_id_saved = lkst.current_maskset_id;
		lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING,
					   lkst.current_maskset_id);
		LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);
	} else {
		/* after dumping, recover previous maskset */
		lkst_hook_maskset_xchg(maskset_id_saved,
					   LKST_MASKSET_ID_RNOTHING);
		LKST_MASKSET_SET_CURRENT(maskset_id_saved);
	}
	lkst_evhandlerprim_entry_log_static_inline(LKST_ETYPE_LKST_KERNEL_DUMP,
						   LKST_ARGP(dev),
						   LKST_ARG(stat),
						   0, 0);
	return NOTIFY_DONE;
}
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

/*********************************************************************
 * Lkst core functions
 *********************************************************************/
/*
 * This checks read mode of a specified buffer list.
 * If the current process is set to read the buffer list,
 * it returns 0. If not, it returns 1.
 * When the process is not set, that is not allowed to
 * read, seek, create, delete buffers of the buffer list.
 */
static inline int lkst_buffer_rmod_check(int cpu)
{
	if (lkst_cpu[cpu].owner.pid != current->pid) {
		return 1 ;/*(lkst_cpu[cpu].buffer_owner != LKST_OWNER_PID_VOID);*/
	}
	return 0;
}

static inline int lkst_trc_status_core(struct lkst_status_param *param)
{
	int i;
	lkst_maskset *maskset;
	lkst_evhandler_func evhandler;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	param->current_maskset_id = lkst.current_maskset_id;

	param->static_buffer_recid = atomic_read(&lkst.static_buffer.recid);
	param->static_buffer_size = LKST_STATIC_BUFFER_BYTE;

	param->nr_cpus = LKST_NR_CPUS;
	for (i = 0; i < LKST_NR_CPUS; i++) {
		param->buffer_id[i].write = lkst_cpu[i].write_buf;
		param->buffer_id[i].read = lkst_cpu[i].owner.read_buf;
	}

	/*needless to get lock, just ignore to count modifying masksets*/
	param->maskset_num = 0;
	for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
		if (LKST_POINTER_LOAD(lkst.maskset_table[i].masksetp, maskset))
			param->maskset_num++;
	}
	/*needless to get lock, just ignore to count modifying evhandler*/
	param->evhandler_num = 0;
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		if (LKST_POINTER_LOAD(lkst.evhandler_table[i].evhandler,
					  evhandler) ) param->evhandler_num++;
	}
	/* get version */
	param->version.major	= LKST_VERSION_MAJOR;
	param->version.minor	= LKST_VERSION_MINOR;
	param->version.revision = LKST_VERSION_REVISION;
	return 0;
}

/* lkst_trc_start* and lkst_trc_stop* functions are obsolete.*/
static inline int lkst_trc_start_core(void)
{
	lkst_maskset *maskset;
	lkst_maskset_id current_id;

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}

	if (lkst.maskset_id_saved > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Save maskset ID=%d is invalid!\n", lkst.maskset_id_saved);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	if (!LKST_POINTER_LOAD(lkst.maskset_table[lkst.maskset_id_saved].masksetp, maskset)) {
		LKST_DPRINTK("Saved maskset ID=%d has already been deleted!\n", lkst.maskset_id_saved);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	current_id = lkst.current_maskset_id;
	if (current_id != lkst.maskset_id_saved){
		LKST_MASKSET_SET_CURRENT(lkst.maskset_id_saved);
		lkst_hook_maskset_xchg(lkst.current_maskset_id, current_id);
	}

	spin_unlock(&lkst_maskset_lock);
	return 0;
}


static inline int lkst_trc_stop_core(void)
{
	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}

	lkst.maskset_id_saved = lkst.current_maskset_id;

	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING,
				   lkst.current_maskset_id);
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);

	spin_unlock(&lkst_maskset_lock);
	return 0;
}

static inline int lkst_maskset_read_core(struct lkst_maskset_param *param)
{
	int i, retval;
	lkst_maskset *maskset;
	struct lkst_maskset_entry *entry;
	lkst_evhandler_id id;
	size_t size;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	size = param->maskset_size;

	if (param->id == LKST_MASKSET_ID_VOID)
		param->id = lkst.current_maskset_id;

	if (param->id > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Invalid maskset ID=%d\n", param->id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/

	if (!LKST_POINTER_LOAD(lkst.maskset_table[param->id].masksetp,
				   maskset)) {
		LKST_DPRINTK("Maskset(ID=%d) does not exist!\n", param->id);
		retval = -EINVAL;
		goto endof_atomic_region;
	}

	entry = param->maskset->entry;
	param->maskset->len = LKST_MASKSET_TABLE_LEN_MAX;

		size -= LKST_MASKSET_SIZE(0);

	for (i = 0; i < LKST_MASKSET_TABLE_LEN_MAX; i++) {
		id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, i);
		entry->event_type = i;
		entry->id = id;
		size -= sizeof(struct lkst_maskset_entry);
		if (size < sizeof(struct lkst_maskset_entry)){
			param->maskset->len = i + 1;
			break;
		}
		entry++;
	}

	strncpy(param->maskset->name, lkst.maskset_table[param->id].name, (size_t) LKST_MASKSET_NAME_LEN);
	param->maskset->name[LKST_MASKSET_NAME_LEN - 1] = '\0';
	retval = 0;

 endof_atomic_region:
	spin_unlock(&lkst_maskset_lock);

	return retval;
}

/* no-lock operation: you must get maskset lock before calling this!*/
static inline lkst_maskset_id lkst_maskset_get_id_core_nl(const char *name)
{
	/* !NOTE!: please check name != NULL && name[0] != '\0'*/
	lkst_maskset_id id;
	volatile lkst_maskset *maskset;
	for (id = 0; id <= LKST_MASKSET_ID_MAX; id++) {
		if (LKST_POINTER_LOAD(lkst.maskset_table[id].masksetp,
					  maskset)) {
			if (strncmp(name,
					lkst.maskset_table[id].name,
					LKST_MASKSET_NAME_LEN) == 0){
				return id;
			}
		}
	}
	return LKST_MASKSET_ID_VOID;
}

static inline \
int lkst_maskset_validate_evhandler_nl(struct lkst_maskset_param *param)
{
	int i;
	lkst_evhandler_id id;
	lkst_evhandler_func evhandler;
	for (i = 0; i < param->maskset->len; i++ ) {
		int type = param->maskset->entry[i].event_type;
		id = param->maskset->entry[i].id;

		if (type > LKST_ETYPE_MAX || type < 0) {
			LKST_DPRINTK("Invalid event_type=%d\n", type);
			return -EINVAL;
		}

		if (id > LKST_EVHANDLER_ID_MAX &&
			id != LKST_EVHANDLER_ID_VOID) {
			LKST_DPRINTK("Invalid handler ID=%d\n", id);
			return -EINVAL;
		}
		if (!LKST_POINTER_LOAD(
				lkst.evhandler_table[id].evhandler,
				evhandler) &&
			id != LKST_EVHANDLER_ID_VOID) {
				LKST_DPRINTK("Handler(ID=%d) does not exist!\n", id);
			return -EINVAL;
		}
	}
	return 0;
}

static inline \
lkst_evhandler_id * lkst_maskset_create_table_inline(void)
{
	int i;
	lkst_evhandler_id * table;
	if (lkst.used_memory_size + LKST_MASKSET_TABLE_BYTE >
		lkst.available_memory_size ) {
		LKST_DPRINTK("Cannot use kernel memory any more!\n");
		return NULL;
	}
	table = (lkst_evhandler_id *)vmalloc(LKST_MASKSET_TABLE_BYTE);
	if (table == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory!\n");
		return NULL;
	}
	lkst.used_memory_size += LKST_MASKSET_TABLE_BYTE;
	for (i = 0; i <= LKST_ETYPE_MAX; i++)
		table[i] = LKST_EVHANDLER_ID_VOID;

	return table;
}

/* !NOTE!: please check (param != NULL && param->maskset != NULL)
 * before calling this.*/
static int lkst_maskset_write_core(struct lkst_maskset_param *param)
{
	int i, ret;
	lkst_maskset *maskset;

	/*get the maskset lock*/
	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_outer_region:*/
	if ( param->maskset->name[0] == '\0' ) {
		if( param->id == LKST_MASKSET_ID_VOID ||
			! LKST_MASKSET_TABLE_LOAD(param->id, maskset)) {
			ret = -EINVAL;
			goto endof_outer_region;
		}
	} else {
		lkst_maskset_id mid;
		mid = lkst_maskset_get_id_core_nl(param->maskset->name);
		if ( mid == LKST_MASKSET_ID_VOID ) {
			/* name is unique */
			if(param->id == LKST_MASKSET_ID_VOID) {
				/* search empty entry */
				for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
					if (!LKST_MASKSET_TABLE_LOAD(i,
									 maskset))
						break;
				}
				if (i == LKST_MASKSET_TBL_MAX) {
					LKST_DPRINTK("Cannot register "
							 "maskset any more!\n");
					ret = -ENOSPC;
					goto endof_outer_region;
				}
				param->id = i;
			}
		} else {/* name is not unique */
			if ( param->id == LKST_MASKSET_ID_VOID ) {
				param->id = mid;
			} else if ( mid != param->id ){
				LKST_DPRINTK("Specified ID does not "
						 "associate with %s!\n",
						 param->maskset->name);
				ret = -EINVAL;
				goto endof_outer_region;
			}/*mid == param->id*/
		}
	}

	/* check whether specified id is reserved*/
	if (param->id <= LKST_MASKSET_ID_RDEFAULT ||
		param->id > LKST_MASKSET_ID_MAX ) {
		LKST_DPRINTK("Invalid number of maskset ID=%d\n", param->id);
		ret = -EINVAL;
		goto endof_outer_region;
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get evhandler lock.\n");
		ret = -EAGAIN;
		goto endof_outer_region;
	}
 /*startof_inner_region:*/
	/* check whether all of specified evhandlers are valid*/
	if ( (ret = lkst_maskset_validate_evhandler_nl(param)) < 0 ) {
		goto endof_inner_region;
	}
	ret = 0;

	if (lkst.maskset_table[param->id].masksetp == NULL) {
		/* new maskset-table allocate */
		maskset = (lkst_maskset *)lkst_maskset_create_table_inline();
		if (!maskset) {
			ret = -ENOMEM;
			goto endof_inner_region;
		}
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, param->id,
						 param->maskset->name, maskset);
	} else {
		/* table will be overwritten */
		if (!LKST_MASKSET_TABLE_LOAD(param->id, maskset)) {
			LKST_DPRINTK("Maskset(ID=%d) is collapsed!\n", param->id);
			ret = -EBUSY;
			goto endof_inner_region;
		}
		if (param->maskset->name[0] != '\0') {
			LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
							 param->id,
							 param->maskset->name,
							 maskset);
		}
	}

	if(param->id == lkst.current_maskset_id) {
		for (i = 0; i < param->maskset->len; i++) {
			lkst_event_type etype =
				param->maskset->entry[i].event_type;
			if((LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype) !=
				param->maskset->entry[i].id)&&
			   (etype < LKST_MASKSET_TBL_MAX)) {
				lkst_hook_set_exit(etype,
						   param->maskset->entry[i].id);
			}
		}
	}

	/* main */
	for (i = 0; i < param->maskset->len; i++)
		if (LKST_ETYPE_VERIFY(param->maskset->entry[i].event_type))
			LKST_MASKSET_SET_EVHANDLER_ID(maskset, 
						      param->maskset->entry[i].event_type,
						      param->maskset->entry[i].id);
 endof_inner_region:
	spin_unlock(&lkst_evhandler_lock);
 endof_outer_region:
	spin_unlock(&lkst_maskset_lock);
	return ret;
}


static inline int lkst_maskset_xchg_core(lkst_maskset_id * id)
{
	int	ret;
	/* !NOTE!: please check (id != NULL) before calling this.*/
#ifdef LKST_DEBUG_MESSAGE
	lkst_maskset_id old_id = *id;
#endif

	ret = lkst_evhandlerprim_maskset_xchg_inline(id);

#ifdef LKST_DEBUG_MESSAGE
	switch (ret) {
	case LKST_MASKSET_XCHG_ERRNO_SUCCESS:
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_WCPT:
		LKST_DPRINTK("Current maskset(ID=%d) is collapsed!\n", old_id);
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_WCID:
		LKST_DPRINTK("Invalid ID=%d of current maskset!\n", old_id);
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_ECPT:
		LKST_DPRINTK("Current maskset(id=%d) is collapsed!\n", old_id);
		return -EBUSY;
	case LKST_MASKSET_XCHG_ERRNO_ENID:
		LKST_DPRINTK("Invalid maskset ID=%d\n", *id);
		return -EINVAL;
	case LKST_MASKSET_XCHG_ERRNO_ENPT:
		LKST_DPRINTK("Specified maskset(ID=%d) does not exist!\n", *id);
		return -EINVAL;
	case LKST_MASKSET_XCHG_ERRNO_ESTAT:
		LKST_DPRINTK("Cannot change maskset when LKST is not running!\n");
		return -EBUSY;
	case LKST_MASKSET_XCHG_ERRNO_ELOCK:
		LKST_DPRINTK("Cannot get maskset lock!\n");
		return -EAGAIN;
	default:
		LKST_DPRINTK("Unknown error occured, old/new maskset ID=%d/%d.\n", old_id, *id);
		return -EINVAL;
	}
#else /*!LKST_DEBUG_MESSAGE*/
	return ret;
#endif
}


static inline int lkst_maskset_list_core(struct lkst_maskset_listparam *lp)
{
	int i, cnt, cnt_max;
	lkst_maskset *maskset;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	if(lp->listent_size < LKST_MASKSET_LISTENT_SIZE(1) ||
	   lp->listent_size > LKST_MASKSET_LISTENT_SIZE(LKST_MASKSET_TBL_MAX)) {
		LKST_DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_maskset_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_MASKSET_ID_VOID;
		lp->listent[i].name[0]='\0';
		lp->listent[i].len = 0;
	}

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/
	for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
		if (LKST_POINTER_LOAD(
				lkst.maskset_table[i].masksetp,
				maskset)) {
			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name, lkst.maskset_table[i].name);

			lp->listent[cnt].len = LKST_MASKSET_TABLE_LEN_MAX;
			cnt++;
			if(cnt >= cnt_max) break;
		}
	}

	lp->current_id = lkst.current_maskset_id;

 /*endof_atomic_region:*/
	spin_unlock(&lkst_maskset_lock);

	return 0;
}


static inline int lkst_maskset_delete_core(lkst_maskset_id id)
{
	int ret;
	lkst_maskset *maskset;

	if (id == LKST_MASKSET_ID_RALL ||
		id == LKST_MASKSET_ID_RNOTHING ||
		id == LKST_MASKSET_ID_RDEFAULT) {
		LKST_DPRINTK("Cannot delete special maskset!\n");
		return -EINVAL;
	}

	if (id > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Invalid maskset ID=%d\n", id);
		return -EINVAL;
	}


	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
	ret = 0;
 /*startof_atomic_region:*/
	if (id == lkst.maskset_id_saved) {
		LKST_DPRINTK("Specified maskset(ID=%d) is currently saved!\n Change to DEFAULT maskset.", id);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}
	if (id == lkst.current_maskset_id) {
		LKST_DPRINTK("Specified maskset(ID=%d) is currently used!\n", id);
		ret = -EBUSY;
	} else {
		if (!LKST_POINTER_LOAD(lkst.maskset_table[id].masksetp,
					   maskset)) {
			LKST_DPRINTK("Maskset(ID=%d) does not exist!\n", id);
			ret = -EINVAL;
		} else {
			LKST_POINTER_STORE(lkst.maskset_table[id].masksetp,
					   NULL);
			lkst.used_memory_size -= LKST_MASKSET_TABLE_BYTE;
			vfree(maskset);
		}
	}
 /*endof_atomic_region:*/
	spin_unlock(&lkst_maskset_lock);

	return ret;
}

static inline int
lkst_evhandler_list_core(struct lkst_evhandler_listparam *lp)
{
	int i, cnt, cnt_max;
	lkst_evhandler_func evhandler;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	if(lp->listent_size < sizeof(struct lkst_evhandler_listent) ||
	   lp->listent_size > LKST_EVHANDLER_LISTENT_SIZE(LKST_EVHANDLER_TBL_MAX)) {
		LKST_DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_evhandler_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_EVHANDLER_ID_VOID;
		lp->listent[i].name[0] = '\0';
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		if (LKST_POINTER_LOAD(
				lkst.evhandler_table[i].evhandler,
				evhandler)) {
			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name, lkst.evhandler_table[i].name);

			cnt++;
			if(cnt > cnt_max)
				break;
		}
	}
	spin_unlock(&lkst_evhandler_lock);

	return 0;
}

static inline int
lkst_evhandler_ctrl_core(lkst_evhandler_id id,
			 void * buf, size_t bufsize,
			 int * retval)
{
	lkst_evhandler_ctrl_func evhandler_ctrl;
	if (id > LKST_EVHANDLER_ID_MAX) {
		LKST_DPRINTK("Invalid handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!LKST_POINTER_LOAD(lkst.evhandler_table[id].evhandler_ctrl,
				   evhandler_ctrl)) {
		LKST_DPRINTK("Ctrl-function(ID=%d) does not exist!\n", id);
		return -EINVAL;
	}

	*retval = (*evhandler_ctrl) (buf, bufsize);
	return 0;
}

/*
 * lkst_evhandler_register_core_nl
 * return value :
 *	success 0
 *	fail	errno
 * Called from lkst_evhandler_register, lkst_evhandler_unregister_core,
 * lkst_init_stage0
 * This function do !NOT! check which name is used.
 * please check it before call.
 * execute without lock.
 */
static inline int
lkst_evhandler_register_core_nl(lkst_evhandler_id id, const char *name,
				lkst_evhandler_func evhandler,
				lkst_evhandler_ctrl_func evhandler_ctrl)
{
	if ( name != NULL ) {
		strncpy(lkst.evhandler_table[id].name, name,
			(size_t) LKST_EVHANDLER_NAME_LEN);
		lkst.evhandler_table[id].name[LKST_EVHANDLER_NAME_LEN - 1] = '\0';
	} else
		lkst.evhandler_table[id].name[0] = '\0';

	/*register*/
	LKST_POINTER_STORE(lkst.evhandler_table[id].evhandler_ctrl,
			   evhandler_ctrl);
	LKST_POINTER_STORE(lkst.evhandler_table[id].evhandler,
			   evhandler);
	return 0;
}

/*
 * lkst_evhandler_unregister_core
 * return value :
 *	success	0
 *	fail	errno
 * Called from lkst_evhandler_unregister, lkst_trc_init_stage0, lkst_cleanup_module
 */
static inline int lkst_evhandler_unregister_core_nl(lkst_evhandler_id id)
{
	return lkst_evhandler_register_core_nl(id, NULL, NULL, NULL);
}

/*
 * lkst_evhandler_get_id_core_nl
 * return value : evhandler_id
 * Called from lkst_evhandler_register,lkst_evhandler_get_id, lkst_evhandler_get_id_ioctl(not implemented)
 */
static inline lkst_evhandler_id
lkst_evhandler_get_id_core_nl(const char *name)
{
	lkst_evhandler_id id;
	/* !NOTE!: please check (name != NULL) && (strlen(name)!=0) before calling.*/
	for ( id = 0; id <= LKST_EVHANDLER_ID_MAX; id++) {
		if (lkst.evhandler_table[id].evhandler != NULL) {
			if (strncmp(name,
					lkst.evhandler_table[id].name,
					LKST_EVHANDLER_NAME_LEN) == 0){
				return id;
			}
		}
	}
	return LKST_EVHANDLER_ID_VOID;
}

/*
 * This function just makes a instance of a buffer.
 * And never sets in the table of buffers.
 */
static inline struct lkst_event_buffer * \
lkst_buffer_create_instance_inline(size_t size, lkst_buffer_id next, int cpu_id)
{
	struct lkst_event_buffer *buffer;
	int i;
	/* align to times of 4k */
	if ( (size & ~LKST_BUFFER_PAGE_MASK) ) size += LKST_BUFFER_PAGE_BYTE;
	size &= LKST_BUFFER_PAGE_MASK;
	if (size < LKST_BUFFER_SIZE_MIN) {
		size = LKST_BUFFER_SIZE_MIN;
	} else if (size > LKST_BUFFER_SIZE_MAX) {
		size = LKST_BUFFER_SIZE_MAX;
	}

	size += LKST_BUFFER_PAGE_BYTE; /* for margin and header */

	/* create new buffer area */
	if (lkst.used_memory_size + size > lkst.available_memory_size ) {
		LKST_DPRINTK("Cannot use kernel memory any more!\n");
		return NULL;
	}
	buffer = (struct lkst_event_buffer *)vmalloc(size);
	if (buffer == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory(size=%d)\n", size);
		return NULL;
	}
	lkst.used_memory_size += size;

	buffer->next = next;
	buffer->rq_next = LKST_BUFFER_ID_VOID;
	buffer->rq_prev = LKST_BUFFER_ID_VOID;

	buffer->size = (size - LKST_BUFFER_HEADER_BYTE)/LKST_EVENT_RECORD_BYTE;
	buffer->baseid = 0;
	buffer->write_offset = -LKST_BUFFER_INIT_POS;
	buffer->cpu = smp_processor_id();
	buffer->cmos_time_flag = lkst.cmos_time_flag;
	buffer->xtime.tv_sec = lkst.xtime.tv_sec;
	buffer->xtime.tv_usec = lkst.xtime.tv_usec;
	strncpy(buffer->eyecatcher, "LKBF", 4);
	buffer->buf_ver = LKST_BUF_VER;

	/* clear recid, cpu_id, and event_type */
	for (i = 0; i < buffer->size; i++) {
		LKST_BUFFER_BODY(buffer)[i].log_recid = 0;
		LKST_BUFFER_BODY(buffer)[i].log_event_type = LKST_ETYPE_VOID;
		LKST_BUFFER_BODY(buffer)[i].processor = cpu_id;
	}
	return buffer;
}

/*
 * This makes a buffer and sets in the table of buffers.
 */
static inline int
lkst_buffer_create_core(lkst_buffer_id id, size_t size, lkst_buffer_id next)
{
	struct lkst_event_buffer *buffer;
	int i;

	buffer = lkst_buffer_create_instance_inline(size, next,
							smp_processor_id());
	if ( buffer == NULL ) return -ENOMEM;

	if (id == LKST_BUFFER_ID_VOID) {
		/* search empty entry */
		for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
			if (LKST_BUFFER_CHECK_NULL(i))
				break;
		}
		if (i == LKST_BUFFER_TBL_MAX) {
			LKST_DPRINTK("Cannot create buffer any more!\n");
			lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
			vfree(buffer);
			return -ENOSPC;
		}
		id = i;
	}else if (! LKST_BUFFER_CHECK_NULL(id)) {
		LKST_DPRINTK("Buffer(ID=%d) already exist!\n", id);
		lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
		vfree(buffer);
		return -EINVAL;
	}

	/* register to buffer_table */ {
		unsigned long flags;
		buffer->id = id;
		local_irq_save(flags);
		LKST_BUFFER_STORE(id, buffer);
		local_irq_restore(flags);
	}
	LKST_DPRINTK("Create a buffer cpu=%d id=%d addr=%p size=%ul\n",
			 smp_processor_id(), id, buffer, size);

	return id;
}

static inline int lkst_buffer_delete_core(lkst_buffer_id id)
{
	struct lkst_event_buffer *buffer = NULL;
	unsigned long flags;
	int ret = 0;

	if (id == LKST_BUFFER_ID_VOID) {
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) does not exist!\n",
				 smp_processor_id(), id);
		return -EINVAL;
	}
	local_irq_save(flags);
	if (lkst_current.write_buf == id) {
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) is currently writing!\n",
				 smp_processor_id(), id);
		ret = -EBUSY;
	} else if ( lkst_current.owner.read_buf == id && lkst_current.reading ){
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) is currently reading!\n",
				 smp_processor_id(), id);
		ret = -EBUSY;
	} else {
		if (!LKST_BUFFER_LOAD(id, buffer)) {
			LKST_DPRINTK("Buffer(cpu=%d ID=%d) does not exist!\n",
					 smp_processor_id(), id);
			ret = -EINVAL;
		} else {
			/* unregister buffer pointer */
			if(buffer->rq_next != LKST_BUFFER_ID_VOID ||
			   lkst_current.rq_tail == id)
				lkst_buffer_dequeue_inline(buffer);
			/* Check whether this buffer is currently using. */
			LKST_BUFFER_STORE(id, NULL);
			if ( lkst_current.owner.read_buf == id ){
				lkst_current.owner.read_buf =
					LKST_BUFFER_ID_VOID;
			}
		}
	}
	local_irq_restore(flags);

	/* free buffer area */
	if (ret == 0) {
		lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
		vfree(buffer);
	}
	LKST_DPRINTK("Delete a Buffer(cpu=%d ID=%d)\n",smp_processor_id(), id);

	return ret;
}

static __inline__ \
int lkst_buffer_list_core(struct lkst_buffer_listparam *lp)
{
	int i, cnt, cnt_max;
	struct lkst_event_buffer *buffer;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	cnt = 0;
	cnt_max = lp->listent_num;

	/* list buffers */
	for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
		if (!LKST_BUFFER_LOAD(i, buffer)) continue;
		if (cnt < cnt_max) {
			lp->listent[cnt].id = buffer->id;
			lp->listent[cnt].next = buffer->next;
			lp->listent[cnt].rq_prev = buffer->rq_prev;
			lp->listent[cnt].rq_next = buffer->rq_next;
			lp->listent[cnt].size = buffer->size*LKST_EVENT_RECORD_BYTE;
			lp->listent[cnt].write_offset = buffer->write_offset;
			lp->listent[cnt].baseid = buffer->baseid;
		}
		cnt++;
	}
	/* record percpu header */
	lp->buffer_num = cnt;
	lp->write_buf = lkst_current.write_buf;
	lp->read_buf = lkst_current.owner.read_buf;
	lp->rq_head = lkst_current.rq_head;
	lp->rq_tail = lkst_current.rq_tail;
	lp->owner = lkst_current.owner.pid;
	lp->read_pos = lkst_current.owner.read_pos;

	return cnt;
}

static int lkst_buffer_set_current_to_owner(int cpu)
{
	unsigned long allowed;

	/* !NOTE!: please check sp != NULL before calling this.*/
	if (lkst_current.owner.pid == current->pid) {
		LKST_DPRINTK("This process(pid=%d) is already "
			     "set to read other buffer list.\n", current->pid);
		return -EBUSY;
	}

	/* CPU binding */
	if( lkst_cpu_migrate_save(&allowed, cpu) < 0) //TODO
		return -EINVAL;

	if (lkst_current.owner.pid != LKST_OWNER_PID_VOID &&
	    lkst_current.owner.pid != current->pid) {
		LKST_DPRINTK("Specified buffer list is already set "
			     "reading mode by pid=%d\n",
			     lkst_current.owner.pid);
		lkst_cpu_migrate_restore(allowed);
		return -EBUSY;
	}

	/* Set owner pid to the buffer list of specified CPU */
	lkst_current.owner.pid = current->pid;
	return 0;
}

static int lkst_buffer_setrmod_core(struct lkst_buffer_srmodparam *sp)
{
	int retval;
	retval = lkst_buffer_set_current_to_owner(sp->cpu);
	if (retval < 0) {
		return retval;
	}

	/* get buffer time stamp */
	sp->tsc = lkst.tsc;
	sp->xtime.tv_sec = lkst.xtime.tv_sec;
	sp->xtime.tv_usec = lkst.xtime.tv_usec;
	sp->cpu_freq = lkst.cpu_freq;
	sp->buf_ver = lkst.buf_ver;
	strncpy(sp->arch, lkst.arch, LKST_ARCH_NAME_LEN);
	sp->endian_big = lkst.endian_big;

	return 0;
}

/*
 * lkst_evhandler_register
 * return value
 *	success 0
 *	fail	errno
 */
int lkst_evhandler_register(lkst_evhandler_id id, const char *name,
				lkst_evhandler_func evhandler,
				lkst_evhandler_ctrl_func evhandler_ctrl)
{
	int ret;
	lkst_evhandler_func tmp;

	if (evhandler == NULL) {
		return -EINVAL;
	}

	if (name == NULL || name[0] == '\0' ){
		LKST_DPRINTK("Invalid evhandler name!\n");
		return -EINVAL;
	}
	if (id <= LKST_EVHANDLER_ID_RESERVED) {
		LKST_DPRINTK("Reserved handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/
	/* check whether the name of event handler is unique.*/
	if ( lkst_evhandler_get_id_core_nl(name) != LKST_EVHANDLER_ID_VOID) {
		LKST_DPRINTK("The name <%s> is already used.\n",name);
		ret = -EINVAL;
		goto endof_atomic_region;
	}
	if (id == LKST_EVHANDLER_ID_VOID) {
		/* Get an unused ID. */
		for (id = LKST_EVHANDLER_ID_RESERVED + 1 ;
			 LKST_POINTER_LOAD(lkst.evhandler_table[id].evhandler,tmp);
			 id++ ) {
			if (id > LKST_EVHANDLER_ID_MAX) {
				LKST_DPRINTK("No handler ID is available.\n");
				ret = -ENOSPC;
				goto endof_atomic_region;
			}
		}
	} else if (LKST_POINTER_LOAD(lkst.evhandler_table[id].evhandler,tmp)) {
		LKST_DPRINTK("Specified handler ID is already used.\n");
		ret = -EINVAL;
		goto endof_atomic_region;
	}

	lkst_evhandler_register_core_nl(id, name,
					evhandler,
					evhandler_ctrl);
	ret = lkst_hook_exit_update(id);

 endof_atomic_region:
	spin_unlock(&lkst_evhandler_lock);
	return ret == 0 ? id : ret;
}

int lkst_evhandler_unregister(lkst_evhandler_id id)
{
	int ret;

	if (id <= LKST_EVHANDLER_ID_RESERVED) {
		LKST_DPRINTK("Reserved handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
	if ( lkst.evhandler_table[id].evhandler == NULL) {
		LKST_DPRINTK("Specified handler(ID:%d) does not exist.\n",id);
		ret = -EINVAL;
	} else {
		lkst_evhandler_unregister_core_nl(id);
		ret = lkst_hook_exit_update(id);
	}
	spin_unlock(&lkst_evhandler_lock);
	return ret;
}


/* kernel trc_status wrap function */
int lkst_trc_status(struct lkst_status_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) {
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	return lkst_trc_status_core(param);
}

/* kernel trc_start wrap function */
int lkst_trc_start()
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_start_core();
}

/* kernel trc_stop wrap function */
int lkst_trc_stop()
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_stop_core();
}

/* kernel maskset_read wrap function */
int lkst_maskset_read(struct lkst_maskset_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) {
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_read_core(param);
}

/* kernel maskset_write wrap function */
int lkst_maskset_write(struct lkst_maskset_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL ) {
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}
	if (param->maskset == NULL ) {
		printk(KERN_ERR "Warning : argument param->maskset is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_write_core(param);
}

/* kernel maskset_get_id wrap function */
int lkst_maskset_get_id(const char *name)
{
	int ret;
	/* this might be called from evhandler.
	 * thus don't check whether caller process has admin-privilege.*/
	if (name == NULL || name[0] == '\0') {
		return -EINVAL;
	}
	if (!spin_trylock(&lkst_maskset_lock)) {
		return -EAGAIN;
	} else {
		ret = lkst_maskset_get_id_core_nl(name);
		spin_unlock(&lkst_maskset_lock);
	}
	return ret;
}

/* kernel maskset_xchg wrap function */
int lkst_maskset_xchg(lkst_maskset_id * id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (id == NULL) {
		printk(KERN_ERR "Warning : argument id is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_xchg_core(id);
}

/* kernel maskset_list wrap function */
int lkst_maskset_list(struct lkst_maskset_listparam *lp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lp == NULL) {
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_list_core(lp);
}

/* kernel maskset_delete wrap function */
int lkst_maskset_delete(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_delete_core(id);
}

/* kernel evhandler_list wrap function */
int lkst_evhandler_list(struct lkst_evhandler_listparam *lp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lp == NULL) {
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	return lkst_evhandler_list_core(lp);
}

/* kernel evhandler_ctrl wrap function */
int lkst_evhandler_ctrl(struct lkst_evhandler_ctrl_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL)
		return -EINVAL;

	return lkst_evhandler_ctrl_core(param->id, param->buf, param->bufsize,
					&param->ret);
}

/* kernel evhandler_get_id wrap function */
int lkst_evhandler_get_id(const char *name)
{
	int ret;
	/* this might be called from evhandler.
	 * thus don't check whether caller process has admin-privilege.*/

	if ( name == NULL || name[0] == '\0' ) {
		return -EINVAL;
	}
	if (!spin_trylock(&lkst_evhandler_lock)) {
		return -EAGAIN;
	} else {
		ret =lkst_evhandler_get_id_core_nl(name);
		spin_unlock(&lkst_evhandler_lock);
		return ret;
	}
}

/*
 * function lkst_buffer_jump
 * A primitive part of an Event-handler. This shift the writing buffer to
 * the specified buffer.
 *
 *	Arguments
 *	  next:		ID of the distination buffer.
 *
 *	returns
 *	  errno:		error code
 *
 *	This will be called from event handlers those are invoked by any processes.
 *	So, we can't restrict any unpermitted accesses.
 */
int lkst_buffer_jump(lkst_buffer_id next)
{
	return lkst_buffer_jump_core(next);
}

/*
 * function lkst_evhandlerprim_buffer_shift
 * A primitive part of an Event-handler. This shift the writing buffer to
 * the next buffer.
 *
 *	Arguments
 *	  NONE
 *
 *	returns
 *	  errno:		error code
 *
 *	This will be called from event handlers those are invoked by any processes.
 *	So, we can't restrict any unpermitted accesses.
 */
int lkst_buffer_shift(void)
{
	return lkst_buffer_shift_core();
}

/* kernel buffer_read wrap function */
int lkst_buffer_read(struct lkst_log_buffer *lbuffer)
{
	int ret;
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lbuffer == NULL) {
		printk(KERN_ERR "Warning : argument lbuffer is NULL!\n");
		return -EINVAL;
	}

	lkst_current.reading++; /*read flag up*/
	ret = lkst_buffer_read_core(lbuffer);
	lkst_current.reading--; /*read flag down*/
	return ret ;
}

/* kernel buffer_create wrap function */
int lkst_buffer_create(lkst_buffer_id id, size_t size, lkst_buffer_id next)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_create_core(id, size, next);
}

/* kernel buffer_delete wrap function */
int lkst_buffer_delete(lkst_buffer_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_delete_core(id);
}

/* dump to printk (It might failed to read all dumped text ... TOO LONG!)*/
void lkst_buffer_dump_current_txt(void)
{
	int i;
	int cpu;
	lkst_maskset_id id;
	int err_f;
	struct lkst_event_record *buff_body;
	unsigned int *pArgs;
	lkst_buffer_id buffer_id;
	struct lkst_event_buffer *buffer;
	int	event_type;
	char	*event_type_name;

	/* stop tracing */
	id = LKST_MASKSET_ID_RNOTHING;
	err_f = lkst_evhandlerprim_maskset_xchg(&id);

/* TODO */
	for (cpu = 0; cpu < LKST_CPU_MAX; cpu++) {
		if (!lkst_cpu_is_online(cpu)) continue;
		buffer_id =lkst_cpu[cpu].write_buf;
		if (!((buffer_id <= LKST_BUFFER_ID_MAX) &&
			  (LKST_CURRENT_BUFFER_LOAD(buffer)) ))
			continue;

		for (i = 0; i < buffer->size; i++) {
			buff_body = &(LKST_BUFFER_BODY(buffer)[i]);
			event_type = buff_body->log_event_type;

			/* Stop dump if buffer pointer
			 *	reaches unused entry. */
			if (!LKST_ETYPE_VERIFY(event_type)) continue;

			event_type_name = "";//TODO
			printk("0x%08x: " LKST_FMT_TSC " [mc], \\\n"
				   " type: %25s(0x%04x), pid: %010u, "
				   "cpu: %03u \\\n",
				   buff_body->log_recid,
				   buff_body->log_time,
				   event_type_name, event_type,
				   buff_body->log_pid,
				   cpu);

			pArgs = (unsigned int *) &buff_body->log_arg1;
			printk("  0x%08x 0x%08x 0x%08x 0x%08x \\\n"
				   "  0x%08x 0x%08x 0x%08x 0x%08x \\\n\n",
				   pArgs[0], pArgs[1], pArgs[2], pArgs[3],
				   pArgs[4], pArgs[5], pArgs[6], pArgs[7]);

		}	/* i */
	}		/* cpu */

	/* restart tracing */
	printk("id: %010u\n", id);
	if (!err_f)
		lkst_evhandlerprim_maskset_xchg(&id);

	return ;
}

/*********************************************************************
 * ioctl functions (static *CLOSED* functions)
 ********************************************************************/

/* ioctl trc_status wrap function */
static int lkst_trc_status_ioctl(struct lkst_status_param *param_user)
{
	int retval = 0;
	struct lkst_status_param param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	retval = lkst_trc_status(&param);

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_status_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl trc_start wrap function */
static int lkst_trc_start_ioctl(void)
{
/*	if (!capable(CAP_SYS_ADMIN)) */
/*		return -EPERM; */

	return lkst_trc_start();
}

/* ioctl trc_stop wrap function */
static int lkst_trc_stop_ioctl(void)
{
/*	if (!capable(CAP_SYS_ADMIN)) */
/*		return -EPERM; */

	return lkst_trc_stop();
}

/* ioctl maskset_read wrap function */
static int lkst_maskset_read_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset;
	struct lkst_maskset_body *user_maskset;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 ||
	    verify_area(VERIFY_WRITE, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Invalid argument(maskset/maskset_size)\n");
		return -EINVAL;
	}

	user_maskset = param.maskset;
	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.maskset_size);
		return -ENOMEM;
	}

	param.maskset = maskset;
	retval = lkst_maskset_read(&param);

	if (copy_to_user(user_maskset, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_maskset);
		return -EINVAL;
	}

	param.maskset = user_maskset;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_write wrap function */
static int lkst_maskset_write_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset, *save_user_maskset;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 ||
	    verify_area(VERIFY_READ, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Invalid argument(maskset/maskset_size)\n");
		return -EINVAL;
	}

	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.maskset_size);
		return -ENOMEM;
	}

	if (copy_from_user(maskset, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param.maskset);
		return -EINVAL;
	}

	save_user_maskset = param.maskset;
	param.maskset = maskset;

	retval = lkst_maskset_write(&param);

	param.maskset = save_user_maskset;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_set wrap function */
static int lkst_maskset_set_ioctl(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_xchg(&id);
}

/* ioctl maskset_list wrap function */
static int lkst_maskset_list_ioctl(struct lkst_maskset_listparam *param_user)
{
	int retval = 0;
	struct lkst_maskset_listparam param;
	struct lkst_maskset_listent *listent;
	struct lkst_maskset_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_maskset_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 ||
	    verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		LKST_DPRINTK("Invalid argument(listent/listent_size)\n");
		return -EINVAL;
	}

	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_maskset_list(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_listent);
		return -EINVAL;
	}

	param.listent = user_listent;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_listparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl maskset_delete wrap function */
static int lkst_maskset_delete_ioctl(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_delete(id);
}

/* ioctl maskset_get_id wrap function */
static int lkst_maskset_get_id_ioctl(struct lkst_maskset_getparam *param_user)
{
	struct lkst_maskset_getparam param;
	int ret;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_maskset_getparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	if (param.name[0] == '\0') {
		return -EINVAL;
	}

	ret = lkst_maskset_get_id(param.name);
	if ( ret < 0 )
		return ret;

	param.id = ret;
	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_getparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	return 0;
}


/* ioctl evhandler_list wrap function */
static int lkst_evhandler_list_ioctl(struct lkst_evhandler_listparam *param_user)
{
	int retval = 0;
	struct lkst_evhandler_listparam param;
	struct lkst_evhandler_listent *listent;
	struct lkst_evhandler_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_evhandler_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 ||
	    verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		LKST_DPRINTK("Invalid argument(listent/listent_size)\n");
		return -EINVAL;
	}

	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_evhandler_list(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_listent);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl evhandler_ctrl wrap function */
static int lkst_evhandler_ctrl_ioctl(struct lkst_evhandler_ctrl_param *param_user)
{
	int retval = 0;
	struct lkst_evhandler_ctrl_param param;
	char *buf = NULL;
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_evhandler_ctrl_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.bufsize < 0 ||
	    verify_area(VERIFY_READ, param.buf, param.bufsize)) {
		LKST_DPRINTK("Invalid argument(buf/bufsize)\n");
		return -EINVAL;
	}

	buf = param.buf; /* save */
	if (param.bufsize != 0) {
		param.buf = vmalloc(param.bufsize);
		if (param.buf == NULL) {
			LKST_DPRINTK("Cannot allocate kernel memory : "
				     "size=%d\n", param.bufsize);
			return -ENOMEM;
		}
		if (copy_from_user(param.buf, buf, param.bufsize)) {
			LKST_DPRINTK("Reading fail at user memory address %p\n"
				     , buf);
			vfree(param.buf);
			return -EINVAL;
		}
	}
	retval = lkst_evhandler_ctrl(&param);

	/* copy return value */
	{ void *ptmp = param.buf; param.buf = buf; buf = ptmp;} /*restore*/
	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_evhandler_ctrl_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		vfree(buf);
		return -EINVAL;
	}
	/* copy contents of arguments */
	if (param.bufsize != 0) {
		if (copy_to_user(param.buf, buf, param.bufsize) ) {
			LKST_DPRINTK("Writing fail at user memory address %p\n",
				     param.buf);
			retval = -EINVAL;
		}
		vfree(buf);
	}
	return retval;
}

/* ioctl evhandler_get_id wrap function */
static int lkst_evhandler_get_id_ioctl(struct lkst_evhandler_getparam *param_user)
{
	struct lkst_evhandler_getparam param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_evhandler_getparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	if (param.name[0] == '\0') {
		return -EINVAL;
	}

	param.id = lkst_evhandler_get_id(param.name);

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_evhandler_getparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	return 0;
}

/* ioctl buffer_read wrap function */
static int lkst_buffer_read_ioctl(struct lkst_log_buffer *lbuffer_user)
{
	int retval = 0;
	struct lkst_log_buffer lbuffer;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&lbuffer, lbuffer_user,
			   sizeof(struct lkst_log_buffer))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     lbuffer_user);
		return -EINVAL;
	}
	if (lbuffer.read_size <= 0 ||
	    verify_area(VERIFY_WRITE, lbuffer.buffer, lbuffer.read_size )) {
		LKST_DPRINTK("Invalid argument(buffer/read_size)\n");
		return -EINVAL;
	}

	if ( lbuffer.id == LKST_BUFFER_ID_VOID && lbuffer.cpu != -1 ){
		LKST_DPRINTK("Invalid argument(buffer id)\n");
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if( lbuffer.id != LKST_BUFFER_ID_VOID ) {
			if ( lkst_cpu_migrate_save(&allowed, lbuffer.cpu) < 0)
				return -EINVAL;
		}

		lkst_current.reading++;/*read flag up*/
		retval = lkst_buffer_read_one_to_user_core(
				lbuffer.id, lbuffer.buffer,
				lbuffer.read_size/LKST_LOG_RECORD_BYTE);
		if (retval >= 0)
			lbuffer.result_read_size = retval*LKST_LOG_RECORD_BYTE;
		lkst_current.reading--;/*read flag down*/

		if( lbuffer.id != LKST_BUFFER_ID_VOID ) {
			lkst_cpu_migrate_restore(allowed);
		}
	}
	lkst_log_header_read_inline(&lbuffer);
	if (copy_to_user(lbuffer_user, &lbuffer,
			 sizeof(struct lkst_log_buffer))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     lbuffer_user);
		return -EINVAL;
	}
	return retval;
}

/* ioctl buffer_create wrap function */
static int lkst_buffer_create_ioctl(struct lkst_buffer_param *param_user)
{
	int retval = 0;
	struct lkst_buffer_param param;

	if (!capable(CAP_SYS_ADMIN))
		//return -EPERM;
		return -101;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_buffer_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		//return -EINVAL;
		return -102;
	}

	/*migrate cpu*/{
		unsigned long allowed;

		LKST_DPRINTK("Migrating cpu=%d...",param.cpu);
		if ( lkst_cpu_migrate_save(&allowed, param.cpu) < 0) {
			LKST_DPRINTK("failed!\n");
//			return -EINVAL;
			return -103;
		}
		LKST_DPRINTK("OK(%d)\n",smp_processor_id());

		retval = lkst_buffer_create(param.id, param.size,
					    param.next);
		if (retval >= 0 && retval <= LKST_BUFFER_ID_MAX) {
			struct lkst_event_buffer * buffer;
			if (!LKST_BUFFER_LOAD(retval, buffer))
				LKST_ENCOUNTER_A_BUG;//TODO
			param.id = (lkst_buffer_id)retval;
			param.result_size= buffer->size*LKST_EVENT_RECORD_BYTE;
		}
		lkst_cpu_migrate_restore(allowed);
	}

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_buffer_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
//		return -EINVAL;
				return -104;
	}

	return retval;
}

/* ioctl buffer_delete wrap function */
static int lkst_buffer_delete_ioctl(struct lkst_buffer_delparam *user_param)
{
	struct lkst_buffer_delparam param;
	int retval;
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_delparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_delete(param.id);
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_shift wrap function */
static int lkst_buffer_shift_ioctl(int cpu)
{
	int retval;
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_shift();
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_jump wrap function */
static int lkst_buffer_jump_ioctl(struct lkst_buffer_jumpparam *user_param)
{
	struct lkst_buffer_jumpparam param;
	int retval;
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_jumpparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_jump(param.dest);
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_list wrap function */
static int lkst_buffer_list_ioctl(struct lkst_buffer_listparam *user_param)
{
	int retval = 0;
	struct lkst_buffer_listparam param;
	struct lkst_buffer_listent *listent;
	struct lkst_buffer_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	if (param.listent_num < 0 ){
		LKST_DPRINTK("Invalid argument(listent_num)\n");
		return -EINVAL;
	}

	if (param.listent_num > 0) {
		if ( verify_area(VERIFY_WRITE, param.listent,
				 param.listent_num*
				 sizeof(struct lkst_buffer_listent)) ){
			LKST_DPRINTK("Invalid argument(listent)\n");
			return -EINVAL;
		}
		user_listent = param.listent;
		listent = vmalloc(param.listent_num*sizeof(struct lkst_buffer_listent));
		if (listent == NULL) {
			LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
				     param.listent_num*sizeof(struct lkst_buffer_listent));
			return -ENOMEM;
		}
	} else {
		user_listent = NULL;
		listent = NULL;
	}

	/*migrate cpu*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;

		param.listent = listent;
		retval = lkst_buffer_list_core(&param);

		lkst_cpu_migrate_restore(allowed);
	}
	param.listent = user_listent;
	if (copy_to_user(user_param, &param,
			 sizeof(struct lkst_buffer_listparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	if (param.listent_num > 0) {
		if (param.listent_num < param.buffer_num)
			param.buffer_num = param.listent_num;
		if (copy_to_user(user_listent, listent,
				 param.buffer_num*sizeof(struct lkst_buffer_listent))){
			LKST_DPRINTK("Writing fail at user memory address %p\n",
				     user_listent);
			return -EINVAL;
		}
		vfree(listent);
	}

	return retval;
}

/* ioctl buffer_link function */
static int lkst_buffer_link_ioctl(struct lkst_buffer_linkparam *user_param)
{
	struct lkst_buffer_linkparam param;
	int ret = 0;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	/* get user parameter */
	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_linkparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}

	if (param.id == param.next) {
		LKST_DPRINTK("Source and Distinaion are same buffer.\n");
		return -EINVAL;
	}

	/*cpu migration*/{
		unsigned long allowed;
		struct lkst_event_buffer * buffer;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		if ( LKST_BUFFER_LOAD(param.id, buffer)) {
			buffer->next = param.next;
		} else {
			ret = -EINVAL;
		}
		lkst_cpu_migrate_restore(allowed);
	}
	return ret;
}

/* ioctl buffer_setrmod wrap function */
static int lkst_buffer_setrmod_ioctl(struct lkst_buffer_srmodparam *param_user)
{
	int retval = 0;
	struct lkst_buffer_srmodparam param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_buffer_srmodparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	retval = lkst_buffer_setrmod_core(&param);

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_buffer_srmodparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl entry_log wrap function */
static int lkst_entry_log_ioctl(struct lkst_entry_args *arg_user)
{
	struct lkst_entry_args arg;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&arg, arg_user, sizeof(struct lkst_entry_args))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     arg_user);
		return -EINVAL;
	}

	(void) lkst_entry_log(NULL, arg.event_type,
			      arg.arg1, arg.arg2, arg.arg3, arg.arg4);

	return (0);
}

/* LKST device open function */
static int lkst_open(struct inode *inode, struct file *filp)
{
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
	int cpu = MINOR(inode->i_rdev) - 1;
	int retval;
	/* Instead of using setrmod, you can open /dev/lkst0-31
	 * those minor numbers are from 1 to 32, to read buffers
	 * of each cpus.*/
	/* But /dev/lkst remains for backward-compatibility and
	 * general interface (for masksets/event-handlers)*/
	if (cpu != -1 ) {
		if (!lkst_cpu_is_online(cpu))
			return -ENODEV;
		retval = lkst_buffer_set_current_to_owner(cpu);
		if (retval < 0)
			return retval;

		filp->private_data = (void *)((lkst_ptr_t)cpu);
	}else
		filp->private_data = (void *)-1;

#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/

#ifdef MODULE
	/* Increment module reference count */
	MOD_INC_USE_COUNT;
#endif /* MODULE */

	return 0;
}


/* LKST device close function */
static int lkst_close(struct inode *inode, struct file *filp)
{
	int i;
#ifdef MODULE
	/* Decrement module reference count */
	MOD_DEC_USE_COUNT;
#endif /* MODULE */

	/* Clear owner pid of buffer list */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		if (lkst_cpu[i].owner.pid == current->pid) {
			lkst_cpu[i].owner.pid = LKST_OWNER_PID_VOID;
			/* CPU unbinding */
			lkst_cpu_migrate_restore(cpu_online_map);
			break;
		}
	}

	return 0;
}


/* LKST device read function */
static ssize_t lkst_read(struct file *filp, char *buf,
			 size_t count, loff_t *ppos)
{
	ssize_t read_count;
	int retval;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lkst_current.owner.pid != current->pid )
		return -EBUSY;

#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
	if ( filp->private_data != (void *)-1 ) {
		struct lkst_log_buffer_header *head;

		head = (struct lkst_log_buffer_header *) kmalloc (
				sizeof(struct lkst_log_buffer_header),
				GFP_KERNEL);
		if (!head) return -ENOMEM;

		/* setting up the header of log buffer*/
		head->header.log_magic = LOGFILE_MAGIC;
		head->header.log_version = 0;

		head->entry.log_size =	2*sizeof(int) +
			sizeof(char)*LKST_ARCH_NAME_LEN;
		head->entry.log_format = PXLOG_BINARY;

		head->endian_big = lkst.endian_big;
		head->buf_ver = lkst.buf_ver;
		strncpy(head->arch, lkst.arch, LKST_ARCH_NAME_LEN);

		/* copy to user space (header only) */
		copy_to_user(buf, head, sizeof(struct lkst_log_buffer_header));
		kfree(head);

		filp->private_data = (void *)-1;
		return sizeof(head);
	}
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/

	if (count < sizeof(struct lkst_log_record) ||
	    verify_area(VERIFY_WRITE, buf, count)) {
		LKST_DPRINTK("Invalid argument(buf/count)\n");
		return -EINVAL;
	}
	count /= sizeof(struct lkst_log_record);
	read_count = 0;

	lkst_current.reading++;/*read flag up*/
 read_retry:
	retval = lkst_buffer_read_all_to_user_core(
			(struct lkst_log_record *)buf,
			(int)count);
	if (retval > 0) read_count += retval;
	if ((retval > 0 && retval < (int)count) || retval == -EAGAIN) {
		if(!(filp->f_flags & O_NONBLOCK)){
			lkst_current.owner.ready = 0;
			interruptible_sleep_on(&lkst_current.owner.wait);
			if (!signal_pending(current)) {
				buf += retval*sizeof(struct lkst_log_record);
				count -= retval;
				goto read_retry;
			}else {
				/* We cannot rewind the read process.
				 * But, at least, we can send read size
				 * back to user process, instead of -EINTR.*/
			}
		}
	} else if ( retval < 0 ){
		lkst_current.reading--;/*read flag down*/
		return retval;
	}
	lkst_current.reading--;/*read flag down*/

	read_count *= sizeof(struct lkst_log_record);
	return read_count; /* we must return in byte*/
}

/* LKST device ioctl function */
static int lkst_ioctl(struct inode *inode, struct file *filp,
	       unsigned int request, unsigned long arg)
{
	int retval = 0;		/* return value */

	switch (request) {
	case LKST_IOC_TRC_STATUS:
		retval =
		    lkst_trc_status_ioctl((struct lkst_status_param *) arg);
		break;
	case LKST_IOC_TRC_START:
		retval = lkst_trc_start_ioctl();
		break;
	case LKST_IOC_TRC_STOP:
		retval = lkst_trc_stop_ioctl();
		break;
	case LKST_IOC_MASKSET_READ:
		retval =
		    lkst_maskset_read_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_WRITE:
		retval =
		   lkst_maskset_write_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_SET:
		retval = lkst_maskset_set_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_MASKSET_LIST:
		retval =
		    lkst_maskset_list_ioctl((struct lkst_maskset_listparam *) arg);
		break;
	case LKST_IOC_MASKSET_DELETE:
		retval = lkst_maskset_delete_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_EVHANDLER_LIST:
		retval =
		    lkst_evhandler_list_ioctl((struct lkst_evhandler_listparam *) arg);
		break;
	case LKST_IOC_EVHANDLER_CTRL:
		retval =
		    lkst_evhandler_ctrl_ioctl((struct lkst_evhandler_ctrl_param *) arg);
		break;

	case LKST_IOC_BUFFER_READ:
		retval =
		    lkst_buffer_read_ioctl((struct lkst_log_buffer *) arg);
		break;
	case LKST_IOC_BUFFER_CREATE:
		retval =
		    lkst_buffer_create_ioctl((struct lkst_buffer_param *) arg);
		break;
	case LKST_IOC_BUFFER_SHIFT:
		retval = lkst_buffer_shift_ioctl((int)arg);
		break;
	case LKST_IOC_BUFFER_JUMP:
		retval = lkst_buffer_jump_ioctl((struct lkst_buffer_jumpparam *)arg);
		break;
	case LKST_IOC_BUFFER_DELETE:
		retval = lkst_buffer_delete_ioctl((struct lkst_buffer_delparam*) arg);
		break;
	case LKST_IOC_BUFFER_LIST:
		retval = lkst_buffer_list_ioctl((struct lkst_buffer_listparam *) arg);
		break;
	case LKST_IOC_BUFFER_LINK:
		retval = lkst_buffer_link_ioctl((struct lkst_buffer_linkparam *) arg);
		break;
	case LKST_IOC_BUFFER_SETRMOD:
		retval = lkst_buffer_setrmod_ioctl((struct lkst_buffer_srmodparam *) arg);
		break;

	case LKST_IOC_ENTRY_LOG:
		retval =
		    lkst_entry_log_ioctl((struct lkst_entry_args *) arg);
		break;
	case LKST_IOC_MASKSET_GETID:
		retval = lkst_maskset_get_id_ioctl(
				(struct lkst_maskset_getparam *) arg);
		break;
	case LKST_IOC_EVHANDLER_GETID:
		retval = lkst_evhandler_get_id_ioctl(
				(struct lkst_evhandler_getparam *) arg);
		break;
	default:
		LKST_DPRINTK("Invalid command no.%d\n", request);
		return -EINVAL;
	}

	return retval;
}


/*-------------------------------------------------------------------
 * Initialize/cleanup functions
 *-------------------------------------------------------------------*/

/*
 * function lkst_trc_init_stage0
 * Initialize the first part of Kernel State Tracer.
 * This function initializes data not concerned with SMP.
 *
 *	Arguments
 *	  NONE
 *	returns
 *	  NONE
 *
 */
void lkst_trc_init_stage0()
{
	int i;

	if ( lkst_init_flag != LKST_NOT_INIT) LKST_ENCOUNTER_A_BUG;

	/* The following values are initialized when kernel was loaded,
	 *	 lkst_cpu[...].recid = 1;
	 *	 maskset_rnothing[], maskset_rall[], maskset_rdefault[]
	 */

	LKST_DPRINTK("Starting initialization stage0\n");

	/* initialize global header */{

		/* Initialize LKST header time stamp */
		lkst_time_set(0);

		spin_lock_init(&lkst_maskset_lock);
		spin_lock_init(&lkst_evhandler_lock);

		/* Clean up maskset table */
		LKST_POINTER_STORE(lkst.current_masksetp, NULL);
		lkst.current_maskset_id = 0;
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
		for ( i = 0 ;i < LKST_MASKSET_TBL_MAX ; i++) {
			lkst.maskset_table[i].name[0] = '\0';
			LKST_POINTER_STORE(
				lkst.maskset_table[i].masksetp, NULL);
		}

		/* Clean up evhandler table */
		for ( i = 0 ;i < LKST_EVHANDLER_TBL_MAX ; i++) {
			lkst.evhandler_table[i].name[0] = '\0';
			LKST_POINTER_STORE(
				lkst.evhandler_table[i].evhandler, NULL);
			LKST_POINTER_STORE(
				lkst.evhandler_table[i].evhandler_ctrl, NULL);
		}

		/* set current buffer */
		atomic_set(&lkst.static_buffer.recid, 0);
		/* clear recid, event_type */
		for (i = 0; i < LKST_STATIC_BUFFER_COUNT; i++) {
			LKST_STATIC_BUFFER_BODY()[i].log_recid = 0;
			LKST_STATIC_BUFFER_BODY()[i].log_event_type =
				LKST_ETYPE_VOID;
		}
	}

	/* initialize percpu header */
	for (i = 0; i < LKST_NR_CPUS; i++) {
		lkst_cpu[i].cpu = i;
		lkst_cpu[i].owner.pid = LKST_OWNER_PID_VOID;
		init_waitqueue_head(&lkst_cpu[i].owner.wait);
		lkst_cpu[i].owner.read_buf = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_head = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_tail = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].write_buf = LKST_BUFFER_ID_VOID;
	}
#if DEBUG
	if (LKST_SIZEOF_LKST_EVENT_RECORD != sizeof(struct lkst_event_record)) {
		/* Verification of LKST_SIZEOF_LKST_EVENT_RECORD is failed. */
		printk(KERN_ERR "LKST error(init):"
			   " sizeof(struct lkst_event_record) is not equal to LKST_SIZEOF_LKST_EVENT_RECORD.\n");
		return;
	}
#endif /* DEBUG */

	/* registering masksets & evhandlers */{
		lkst_maskset *ms;
		spin_lock(&lkst_evhandler_lock);
		spin_lock(&lkst_maskset_lock);

		/* Setup nothing recording maskset */
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
						 LKST_MASKSET_ID_RNOTHING,
						 LKST_MASKSET_NAME_RNOTHING,
						 maskset_rnothing);
		/* And set that currently using maskset */
		LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);


		/* Initialize Default Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_DEFAULT,
						LKST_EVHANDLER_NAME_DEFAULT,
						lkst_evhandler_default,
						lkst_evhandler_default_ctrl);

		/* Initialize Buffer Shift Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_BUFFER_SHIFT,
						LKST_EVHANDLER_NAME_BUFFER_SHIFT,
						lkst_evhandler_buffer_shift,
						NULL);

#if DEBUG
		/* Initialize Exception Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_EXCEPTION,
						LKST_EVHANDLER_NAME_EXCEPTION,
						lkst_evhandler_exception,
						NULL);
#endif /* DEBUG */
		/* Initialize Logging static buffer Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_LOG_STATIC,
						LKST_EVHANDLER_NAME_LOG_STATIC,
						lkst_evhandler_log_static,
						NULL);

		/* Setup other special masksets */
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
						 LKST_MASKSET_ID_RALL,
						 LKST_MASKSET_NAME_RALL,
						 maskset_rall);
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
						 LKST_MASKSET_ID_RDEFAULT,
						 LKST_MASKSET_NAME_RDEFAULT,
						 maskset_rdefault);

		/* Clear entries for other masksets */
		for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
			LKST_POINTER_STORE(lkst.maskset_table[i].masksetp,
					   NULL);
		}
		/* Temporarily modify MASKSET_RALL to a maskset log to static buffer*/
		LKST_MASKSET_TABLE_LOAD(lkst_init_maskset, ms);
		for (i = 0; i < LKST_ETYPE_MAX; i++) {
			if (LKST_MASKSET_QUERY_EVHANDLER_ID(ms, i) != LKST_EVHANDLER_ID_VOID)
				maskset_rall[i] = LKST_EVHANDLER_ID_LOG_STATIC;
			else
				maskset_rall[i] = LKST_EVHANDLER_ID_VOID;
		}

		/* Kernelhooks initialization will invoked automatically now. */
		/* Initialize hook table used by lkst */
		lkst_hook_init();

		/* Setup current maskset pointer (stage0 special maskset)*/
		lkst_hook_maskset_xchg(LKST_MASKSET_ID_RALL,
					   lkst.current_maskset_id);
		LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RALL);

		spin_unlock(&lkst_maskset_lock);
		spin_unlock(&lkst_evhandler_lock);
	}
	LKST_DPRINTK("Initialized masksets and evhandlers\n");

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT,
			   LKST_ARG(0x0001), 0, 0, 0);

	/* SysRq key bind */
#ifdef CONFIG_MAGIC_SYSRQ
	register_sysrq_key('v', &sysrq_lkst_dump_buffer_op);
#endif /* CONFIG_MAGIC_SYSRQ */

	/* register Dump notifier (if available)*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	register_dump_notifier(&dump_nb);
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

	/* LKST initialize flag on */
	lkst_init_flag = LKST_INIT_LEVEL_0;

	LKST_DPRINTK("Initialization stage0 finished\n");

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT,
			   LKST_ARG(0x00ff), 0, 0, 0);
}

static int lkst_trc_init_percpu(int cpu)
{
	struct lkst_event_buffer *buffer;
	if ( cpu > LKST_CPU_MAX ) return -EINVAL;
	if ( lkst_cpu[cpu].write_buf != LKST_BUFFER_ID_VOID ) return 0;

	buffer = lkst_buffer_create_instance_inline(lkst_init_buffer_size,
							LKST_BUFFER_ID_VOID,
							cpu);
	if ( buffer == NULL ) {
		LKST_DPRINTK("Can't allocate a buffer for CPU%d\n", cpu);
		return -EINVAL;
	}
	buffer->id = 0;
	LKST_POINTER_STORE(lkst_cpu[cpu].buffer_table[0].ptr,buffer);

	/* set current buffer */
	lkst_cpu[cpu].write_buf = 0;
	/* enqueue current buffer */
	lkst_cpu[cpu].rq_head = 0;
	lkst_cpu[cpu].rq_tail = 0;

	LKST_DPRINTK( "Initialized CPU%d \n", cpu);
	return 0;
}


/*
 * function lkst_trc_init_stage1
 * Initialize the second part of Kernel State Tracer.
 * This function initializes buffers, pointers and
 * data concernd with SMP. Therefore, this should be
 * called after all of the CPUs are initialized.
 *
 *	Arguments
 *	  NONE
 *	returns
 *	  NONE
 *
 */
int lkst_trc_init_stage1()
{
	int i, ret;

	if ( lkst_init_flag != LKST_INIT_LEVEL_0) {
		LKST_DPRINTK("LKST is not initialized yet!\n");
		return -1;
	}

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG(0x0101), 0, 0, 0);

	LKST_DPRINTK( "Starting initialization stage1\n");

	/* Set architecture name */
	strncpy(lkst.arch, system_utsname.machine, 65);

	/* Set cpu frequency */
	lkst.cpu_freq = lkst_get_cpu_khz();
	lkst_time_set(0);

	/* allocate event buffer for all CPUs */
	for (i = 0; i < smp_num_cpus; i++) {
		lkst_trc_init_percpu(i);
	}
	/* get locks */ {
		spin_lock(&lkst_evhandler_lock);
		spin_lock(&lkst_maskset_lock);

		/* Switch to default maskset from log-static-buffer maskset */
		if (lkst_init_maskset != lkst.current_maskset_id) {
			/* Reset current maskset pointer */
			lkst_hook_maskset_xchg((lkst_maskset_id)lkst_init_maskset,
						   lkst.current_maskset_id);
			LKST_MASKSET_SET_CURRENT((lkst_maskset_id)lkst_init_maskset);
		}

		/* Recovery MASKSET_RALL */
		for (i = 0; i < LKST_ETYPE_MAX; i++) {
			maskset_rall[i] = LKST_EVHANDLER_ID_DEFAULT;
		}
		maskset_rall[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_BUFFER_SHIFT;

		if (lkst_init_maskset == lkst.current_maskset_id) {
			/* reset all entries of current maskset */
			while ( lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING,
							   LKST_MASKSET_ID_RALL) == -EBUSY);
			lkst_hook_maskset_xchg(lkst.current_maskset_id,
						   LKST_MASKSET_ID_RNOTHING);
		}

		/*release locks*/
		spin_unlock(&lkst_maskset_lock);
		spin_unlock(&lkst_evhandler_lock);
	}

	/* register LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
	if (devfs_register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_cdev_fops))
		printk(KERN_ERR "LKST warning(init):cannot register LKST *conventional* device!\n");
#else
	if (register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_cdev_fops))
		printk(KERN_ERR "LKST error(init):cannot register LKST device!\n");
#endif

#ifdef CONFIG_DEVFS_FS
	lkst_devfs_handle = devfs_register (NULL, LKST_DEV_NAME,
					    DEVFS_FL_DEFAULT,
					    LKST_DEV_MAJOR, LKST_DEV_MINOR,
					    S_IFCHR | S_IRUSR | S_IWUSR,
					    &lkst_cdev_fops, NULL);
	if ( lkst_devfs_handle == NULL )
		printk(KERN_ERR "LKST error(init):cannot register LKST devfs device!\n");
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
	for ( i=0; i<LKST_NR_CPUS; i++) {/* for Hotplug CPU...*/
		char devname[32];
		sprintf(devname, "%s%d", LKST_DEV_NAME, i);
		lkst_devfs_handle_cpu[i] = devfs_register (NULL, devname,
							   DEVFS_FL_DEFAULT,
							   LKST_DEV_MAJOR, i+1,
							   S_IFCHR | S_IRUSR | S_IWUSR,
							   &lkst_cdev_fops, NULL);
		if ( lkst_devfs_handle_cpu[i] == NULL )
			printk(KERN_ERR "LKST error(init):cannot register LKST devfs device!\n");
	}
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
#endif

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	/* lkst hook (procfs entry)*/
	if ( (ret = lkst_hook_proc_init()) ) {
		LKST_DPRINTK("failed to initialize lkst proc (%d)\n",ret);
		return ret;
	}
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
	/* LKST initialize flag on */
	lkst_init_flag = LKST_INIT_LEVEL_1;

	LKST_DPRINTK("Initialization stage1 finished\n");

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG(0x10ff), 0, 0, 0);

	return 0;
}


/*
 * function lkst_trc_init_main
 * Initialize basic part of Kernel State Tracer.
 * This function is called from /init/main.c. and lkst_init_module
 */
int lkst_trc_init_main()
{
	int ret=0;
#ifndef	CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START
	lkst_trc_init_stage0();
#endif /* !CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START */
	ret = lkst_trc_init_stage1();
	return ret;
}

#ifdef MODULE
static int __init lkst_init_module(void)
{
	size_t size;

	printk(KERN_INFO "LKST: Installing driver.\n");

	/* buffer size check */
	printk(KERN_INFO "LKST: initial event buffer size=" LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size);
	if (lkst_init_buffer_size < LKST_BUFFER_SIZE_MIN) {
		printk(KERN_ERR "LKST warning(init): buffer size(=" LKST_FMT_SZT ") is too small, changed to " LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size, (size_t)LKST_BUFFER_SIZE_MIN);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MIN;
	} else if (lkst_init_buffer_size > LKST_BUFFER_SIZE_MAX) {
		printk(KERN_ERR "LKST warning(init): buffer size(=" LKST_FMT_SZT ") is too large, changed to " LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size, (size_t)LKST_BUFFER_SIZE_MAX);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MAX;
	}
	size = lkst_init_buffer_size;
	if ( size & ~LKST_BUFFER_PAGE_MASK ) size+=LKST_BUFFER_PAGE_BYTE;
	size &= LKST_BUFFER_PAGE_MASK;
	if (lkst_init_buffer_size != size)
		printk(KERN_ERR "LKST warning(init): buffer size must be times of page size, changed to " LKST_FMT_SZT "\n", (size_t)size);
	lkst_init_buffer_size = size;

	/* maskset id check */
	printk(KERN_INFO "LKST: initial maskset id = %d\n", lkst_init_maskset);
	if (lkst_init_maskset < 0 || lkst_init_maskset > 3){
		printk(KERN_ERR "LKST warning(init): initial maskset id is off limited. set default value\n");
		lkst_init_maskset = LKST_MASKSET_ID_RDEFAULT;
	}
	return lkst_trc_init_main();
}

static void __exit lkst_cleanup_module(void)
{
	int i,j;

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	lkst_hook_proc_cleanup();
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/

	/* unregister LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
	if (devfs_unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#else
	if (unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#endif
		printk(KERN_ERR "LKST error :cannot unregister LKST device!\n");

#ifdef CONFIG_DEVFS_FS
	devfs_unregister ( lkst_devfs_handle );
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
	for ( i=0; i<LKST_NR_CPUS; i++) {/* for Hotplug CPU...*/
		devfs_unregister ( lkst_devfs_handle_cpu[i] );
	}
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
#endif

	/* deregister Dump notifier (if available)*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	unregister_dump_notifier(&dump_nb);
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

#ifdef CONFIG_MAGIC_SYSRQ
	unregister_sysrq_key('v', &sysrq_lkst_dump_buffer_op);
#endif /* CONFIG_MAGIC_SYSRQ */

	spin_lock(&lkst_evhandler_lock);
	spin_lock(&lkst_maskset_lock);

	/* Setup current maskset pointer to RNOTHING */
	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING,
				   lkst.current_maskset_id);
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);

	/* GKHI hook cleanup */
	lkst_hook_cleanup();

	/* Cleanup additional maskset pointer & Free */
	for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
		lkst_maskset * masksetp;
		if (LKST_POINTER_LOAD(lkst.maskset_table[i].masksetp,
					  masksetp)) {
			lkst.used_memory_size -= LKST_MASKSET_TABLE_BYTE;
			vfree(masksetp);
			LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, i,
							 "", NULL);
		}
	}
	/* Cleanup default maskset pointer (except RNOTHING)*/
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					 LKST_MASKSET_ID_RALL, "", NULL);
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					 LKST_MASKSET_ID_RDEFAULT, "", NULL);

	/* Cleanup event handler */
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		lkst_evhandler_unregister_core_nl(i);
	}

	/* Cleanup RNOTHING maskset pointer */
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					 LKST_MASKSET_ID_RNOTHING, "", NULL);
	LKST_POINTER_STORE(lkst.current_masksetp, NULL);

	spin_unlock(&lkst_maskset_lock);
	spin_unlock(&lkst_evhandler_lock);

	/* Cleanup event buffer_table & Deallcation event buffer */
	for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
		struct lkst_event_buffer * bufferp;
		for ( j = 0 ; j < LKST_CPU_MAX; j++) {
			if (LKST_POINTER_LOAD(lkst_cpu[j].buffer_table[i].ptr,
						  bufferp) ) {
				lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(bufferp);
				vfree(bufferp);
				LKST_POINTER_STORE(lkst_cpu[j].buffer_table[i].ptr,
						   NULL);
			}
		}
	}

	LKST_DPRINTK_IF(lkst.used_memory_size!=0,
			"WARNING!:there are memory leaks!(%d)", lkst.used_memory_size);
	atomic_set(&lkst.static_buffer.recid, 0);

	/* Cleanup current buffer */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		lkst_cpu[i].rq_head = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_tail = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].write_buf = LKST_BUFFER_ID_VOID;
	}

	/* LKST initialize flag off */
	lkst_init_flag = LKST_NOT_INIT;

	/* deallocation complete */
	printk(KERN_INFO "LKST :driver removed\n");
}

module_init(lkst_init_module);
module_exit(lkst_cleanup_module);
MODULE_AUTHOR("LKST project");
MODULE_DESCRIPTION("Linux Kernel State Trace driver");
MODULE_LICENSE("GPL");
MODULE_PARM(lkst_init_buffer_size, "i");
MODULE_PARM_DESC(lkst_init_buffer_size, "Default size of each buffer(4096-0x200000, in times of 4k)");
MODULE_PARM(lkst_init_maskset, "i");
MODULE_PARM_DESC(lkst_init_maskset, "a maskset id used in starting up.[0-2]");
#endif /* MODULE */
