/*
 lksteh_sysinfo -- record sys-zone information handler
  
 Copyright (C) HITACHI,LTD. 2005
 WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
  
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/lkst_private.h>
#include <linux/kernel.h>
#include "extra_etypes.h"

MODULE_AUTHOR("M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>");
MODULE_DESCRIPTION("");
MODULE_LICENSE("GPL");

/* Declaration of hook header */
LKST_ETYPE_DEF_MODULE(EID_SYSTEM_INFO, INLINE, SYSTEM_INFO, "system information", \
		      "totalram", \
		      "freeram", \
		      "sharedram",\
		      "bufferram");

#include <linux/mm.h>

static void lkst_evhandler_sysinfo(void *phookrec, int event_type,
				  lkst_arg_t arg1, lkst_arg_t arg2,
				  lkst_arg_t arg3, lkst_arg_t arg4)
{
	struct sysinfo i;
	lkst_evhandlerprim_entry_log(event_type, arg1, arg2, arg3, arg4);
	si_meminfo(&i);
	lkst_evhandlerprim_entry_log(LKST_ETYPE_SYSTEM_INFO,
				     LKST_ARG(i.totalram), 
				     LKST_ARG(i.freeram),
				     LKST_ARG(i.sharedram),
				     LKST_ARG(i.bufferram));
}


static int mod_init(void);
static void mod_cleanup(void);
module_init(mod_init);		/*define as initializer*/
module_exit(mod_cleanup);	/*define as terminator*/

static int mod_init(void)
{
	int	retval;
	/* Initialization of hook header */
	retval = lkst_hook_etype_register(&LKST_ETYPE_INFO(SYSTEM_INFO));
	if (retval < 0) {
		printk("lksteh_sysinfo: Error: failed to etype register (%d)\n", 
		       retval);
		goto init_err;
	}
	/* register an event handler */
	retval = lkst_evhandler_register(LKST_EVHANDLER_ID_VOID, "sysinfo", lkst_evhandler_sysinfo, NULL);
	if (retval < 0) {
		printk("lksteh_sysinfo: Error: failed to evhandler register (%d)\n", 
		       retval);
		goto init_err;
	}
	return 0;
	
	init_err:
	mod_cleanup(); /*If occurs an error, invokes terminator.*/
	return retval;
}


static void mod_cleanup(void)
{
	int ret = lkst_evhandler_get_id("sysinfo");
	if (ret != LKST_EVHANDLER_ID_VOID) {
		lkst_evhandler_unregister(ret);
	}
	lkst_hook_etype_unregister(&LKST_ETYPE_INFO(SYSTEM_INFO));
}
