/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2002 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by T.Nakamura <tetsu@crl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _ARCH_I386_LKST_H
#define _ARCH_I386_LKST_H

#ifdef __KERNEL__
# include <linux/types.h>
#else /*!__KERNEL__*/
# include <sys/types.h>
#endif /*!__KERNEL__*/

static inline void * lkst_generic_current_pc(void)
{
	__label__ here;
here:
	return (&&here);
}

#define lkst_current_pc() lkst_generic_current_pc()

#define LKST_BIG_ENDIAN	0
#define LKST_NR_CPUS_MAX 255

typedef unsigned int lkst_ptr_t;
#define LKST_ARGP(ptr)							\
	((lkst_arg_t) ((lkst_ptr_t)(ptr)))

/* buffer size -- must be times of 4k */
#define LKST_BUFFER_SIZE_DEFAULT	(64*1024)		/* 64kBytes*/
#define LKST_BUFFER_SIZE_MAX		(64*1024*1024)		/* 64MBytes*/
#define LKST_BUFFER_SIZE_MIN		(8192)			/* 8kBytes */

typedef	unsigned long	lkst_cpu_freq_t;
typedef	u_int64_t 	lkst_tsc_t;


#ifdef __KERNEL__

	/* This is ugly fix for arch/i386/boot/compressed/misc.c */
#ifdef LKST_SUPPRESS_PIO_TRACE

#define LKST_ENTRY_LOG_PIO_OUT(portaddr,bw,val,calladdr)
#define LKST_ENTRY_LOG_PIO_IN(portaddr,bw,val,calladdr)

#else  /* i.e., !ifdef LKST_SUPPRESS_PIO_TRACE */

#define LKST_ENTRY_LOG_PIO_OUT(portaddr,bw,val,calladdr)\
  LKST_HOOK_INLINE(LKST_ETYPE_O_PORTOUT, LKST_ARG(portaddr),\
		LKST_ARG(bw), LKST_ARG(val), LKST_ARGP(calladdr));
#define LKST_ENTRY_LOG_PIO_IN(portaddr,bw,val,calladdr)\
  LKST_HOOK_INLINE(LKST_ETYPE_O_PORTIN, LKST_ARG(portaddr),\
		LKST_ARG(bw), LKST_ARG(val,0), LKST_ARGP(calladdr));

#endif /* ifdef LKST_SUPPRESS_PIO_TRACE */

/*
 * return offset from specified jiffies in microseconds.
 */
extern unsigned long (*do_gettimeoffset)(void);
#define lkst_get_time_offset(base_jiffies) \
       ((jiffies - base_jiffies)*(1000000L/HZ) + do_gettimeoffset())

#endif /*__KERNEL__*/

#endif /* _ARCH_I386_LKST_H */
