/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _ARCH_IA64_LKST_H
#define _ARCH_IA64_LKST_H

#ifdef __KERNEL__
# include <linux/types.h>
#else /*!__KERNEL__*/
# include <sys/types.h>
#endif /*!__KERNEL__*/
#include <asm/page.h>

#define LKST_BIG_ENDIAN	0
#define LKST_NR_CPUS_MAX 64

typedef unsigned long lkst_ptr_t;
#define LKST_ARGP(ptr)							\
	((lkst_arg_t) ((lkst_ptr_t)(ptr)))

/* buffer size -- must be times of pagesize */
#define LKST_BUFFER_SIZE_DEFAULT	((size_t)(1*1024*1024))		/* 1MBytes*/
#define LKST_BUFFER_SIZE_MAX		((size_t)(1*1024*1024*1024))	/* 1GBytes*/
#define LKST_BUFFER_SIZE_MIN		((size_t)(PAGE_SIZE))		/* 8k/16k/64kBytes */

typedef	unsigned long	lkst_cpu_freq_t;
typedef	u_int64_t 	lkst_tsc_t;


#ifdef __KERNEL__

extern void lkst_ia64_syscall_entry(u_long, u_long, u_long, u_long);
extern void lkst_ia64_syscall_exit(u_long, u_long, u_long, u_long);

/* IA64 specific definition */
#define lkst_current_pc() \
	({ void *pc; asm volatile ("mov %0=ip" : "=r" (pc)); pc; })

	/* This is ugly fix for arch/i386/boot/compressed/misc.c */
#ifdef LKST_SUPPRESS_PIO_TRACE

#define LKST_ENTRY_LOG_PIO_OUT(portaddr,bw,val,calladdr)
#define LKST_ENTRY_LOG_PIO_IN(portaddr,bw,val,calladdr)

#else  /* i.e., !ifdef LKST_SUPPRESS_PIO_TRACE */

#define LKST_ENTRY_LOG_PIO_OUT(portaddr,bw,val,calladdr)	\
	LKST_HOOK_INLINE(LKST_ETYPE_O_PORTOUT,			\
			LKST_ARG(portaddr),			\
			LKST_ARG(bw),				\
			LKST_ARG(val),				\
			LKST_ARG(calladdr));
#define LKST_ENTRY_LOG_PIO_IN(portaddr,bw,val,calladdr)		\
	LKST_HOOK_INLINE(LKST_ETYPE_O_PORTIN,			\
			LKST_ARG(portaddr),			\
			LKST_ARG(bw),				\
			LKST_ARG(val),				\
			LKST_ARG(calladdr));

#endif /* ifdef LKST_SUPPRESS_PIO_TRACE */

//TODO : get offset from efi...
#define lkst_get_time_offset(base_jiffies) (0)

#endif /*__KERNEL__*/

#endif /* _ARCH_IA64_LKST_H */
