/*
 * kerntypes.c
 *
 * Copyright (C) 2000 Tom Morano (tjm@sgi.com) and
 *                    Matt D. Robinson (yakker@alacritech.com)
 *
 * Dummy module that includes headers for all kernel types of interest. 
 * The kernel type information is used by the lcrash utility when 
 * analyzing system crash dumps or the live system. Using the type 
 * information for the running system, rather than kernel header files,
 * makes for a more flexible and robust analysis tool.
 *
 * This source code is released under version 2 of the GNU GPL.
 */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#include <linux/module.h>
#include <linux/autoconf.h>
#include <linux/mm.h>
#include <linux/config.h>
#include <linux/utsname.h>
#include <linux/dump.h>
#include <linux/pci.h>

/* The following includes and structures are needed for SGI's SN
 * platform.
 */
#if defined(CONFIG_IA64_SGI_SN2) || defined(CONFIG_IA64_GENERIC)
#include <linux/types.h>
#include <asm/sn/sgi.h>
#include <linux/devfs_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/sn/invent.h>
#include <asm/sn/hcl.h>
#include <asm/sn/labelcl.h>
#include <asm/sn/nodepda.h>
#include <asm/sn/pci/pciio_private.h>
#include <asm/sn/xtalk/xtalk_private.h>

/** 
 ** Various devfs type definitions lifted from fs/devfs/base.c
 **/

struct directory_type
{
    rwlock_t lock;                   /*  Lock for searching(R)/updating(W)   */
    struct devfs_entry *first;
    struct devfs_entry *last;
    unsigned short num_removable;    /*  Lock for writing but not reading    */
    unsigned char no_more_additions:1;
};

struct file_type
{
    unsigned long size;
};

struct device_type
{
    unsigned short major;
    unsigned short minor;
};

struct fcb_type  /*  File, char, block type  */
{
    void *ops;
    union
    {
        struct file_type file;
        struct device_type device;
    }
    u;
    unsigned char auto_owner:1;
    unsigned char aopen_notify:1;
    unsigned char removable:1;  /*  Belongs in device_type, but save space   */
    unsigned char open:1;       /*  Not entirely correct                     */
    unsigned char autogen:1;    /*  Belongs in device_type, but save space   */
};

struct symlink_type
{
    unsigned int length;         /*  Not including the NULL-termimator       */
    char *linkname;              /*  This is NULL-terminated                 */
};

struct devfs_inode     /*  This structure is for "persistent" inode storage  */
{
    struct dentry *dentry;
    time_t atime;
    time_t mtime;
    time_t ctime;
    unsigned int ino;            /*  Inode number as seen in the VFS         */
    uid_t uid;
    gid_t gid;
};

struct devfs_entry
{
#ifdef CONFIG_DEVFS_DEBUG
    unsigned int magic_number;
#endif
    void *info;
    atomic_t refcount;           /*  When this drops to zero, it's unused    */
    union
    {
        struct directory_type dir;
        struct fcb_type fcb;
        struct symlink_type symlink;
        const char *name;        /*  Only used for (mode == 0)               */
    }
    u;
    struct devfs_entry *prev;    /*  Previous entry in the parent directory  */
    struct devfs_entry *next;    /*  Next entry in the parent directory      */
    struct devfs_entry *parent;  /*  The parent directory                    */
    struct devfs_entry *slave;   /*  Another entry to unregister             */
    struct devfs_inode inode;
    umode_t mode;
    unsigned short namelen;      /*  I think 64k+ filenames are a way off... */
    unsigned char hide:1;
    unsigned char vfs_deletable:1;/*  Whether the VFS may delete the entry   */
    char name[1];                /*  This is just a dummy: the allocated array
                                     is bigger. This is NULL-terminated      */
};
#endif

void
kerntypes_dummy(void)
{
}
